#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Pause action
============================================================================

"""


import time
from dragonfly.actions.action_base import DynStrActionBase


class Pause(DynStrActionBase):
    """
        Pause for the given amount of time.

        The *spec* constructor argument should be a *string* giving the
        time to wait.  It should be given in hundredths of a second.  For
        example, the following code will pause for 20/100s = 0.2
        seconds: ::

            Pause("20").execute()

        The reason the *spec* must be given as a *string* is because it
        can then be used in dynamic value evaluation.  For example, the
        following code determines the time to pause at execution time: ::

            action = Pause("%(time)d")
            data = {"time": 37}
            action.execute(data)

    """

    def _parse_spec(self, spec):
        interval = float(spec) / 100
        return interval

    # pylint: disable=W0221
    # Arguments differ from DynStrActionBase._execute_events().
    def _execute_events(self, interval):
        time.sleep(interval)
        return True



# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import logging
import threading
import time
import unittest

from dragonfly import (ActionBase, CompoundRule, Grammar, Literal,
                       MappingRule, Rule, get_engine, RPCServer,
                       send_rpc_request)


class CapturingHandler(logging.Handler):

    def __init__(self):
        logging.Handler.__init__(self)
        self.records = []

    def emit(self, record):
        self.records.append(record)


class RPCTestCase(unittest.TestCase):
    """
    Tests for RPC server methods.
    """

    @classmethod
    def setUpClass(cls):
        cls.current_response = None

        # Start the server on a different port for the tests.
        cls.server = RPCServer(port=50052)
        cls.server.start()

    @classmethod
    def tearDownClass(cls):
        cls.server.stop()

    def _send_request(self, request_function):
        # Send requests to the server using new threads to emulate requests
        # from a different process.
        request_thread = threading.Thread(target=request_function)
        request_thread.start()

        # Manually process server requests by calling the timer manager's
        # main callback directly.
        while request_thread.is_alive():
            get_engine()._timer_manager.main_callback()
            time.sleep(0.05)

        # Handle errors and results from the request.
        if isinstance(self.current_response, Exception):
            raise self.current_response
        else:
            return self.current_response

    def send_request(self, method, params, id=0):
        def request():
            try:
                self.current_response = self.server.send_request(
                    method, params, id
                )
            except Exception as e:
                self.current_response = e

        return self._send_request(request)

    def test_add_method(self):
        """ Verify that RPC methods can be added and replaced."""
        self.server.add_method(lambda: "bar", "foo")
        response = self.send_request("foo", [])
        self.assertEqual(response, {'jsonrpc': '2.0', 'result': 'bar',
                                    'id': 0})
        self.server.add_method(lambda: "foo", "foo")
        response = self.send_request("foo", [])
        self.assertEqual(response, {'jsonrpc': '2.0', 'result': 'foo',
                                    'id': 0})

    def test_remove_method(self):
        """ Verify that RPC methods can be removed."""
        # The following should not raise an error.
        self.server.remove_method("non_existent")

        # Add a 'bar' method, check that it works.
        self.server.add_method(lambda: "foo", "bar")
        response = self.send_request("bar", [])
        self.assertEqual(response, {'jsonrpc': '2.0', 'result': 'foo',
                                    'id': 0})

        # Remove it and check that an error is returned by the server.
        # 'send_request' will raise a RuntimeError if the server returns an
        # error.
        self.server.remove_method("bar")
        self.assertRaises(RuntimeError,
                          lambda: self.send_request("bar", []))

    def test_errors(self):
        """ Verify that the server handles errors from RPC methods. """
        log_capture = CapturingHandler()
        logger = logging.getLogger("rpc.methods")
        logger.addHandler(log_capture)

        # Add a new RPC method and check if an error is logged and received.
        def error():
            raise RuntimeError("error message")

        self.server.add_method(error, "error")
        try:
            self.assertRaises(RuntimeError,
                              lambda: self.send_request("error", []))
            self.assertEqual(len(log_capture.records), 1)
            log_message = log_capture.records[0].msg
            self.assertTrue("Exception occurred during RPC method" in log_message)
            self.assertTrue("error message" in log_message)
        finally:
            logger.removeHandler(log_capture)

    def test_security_tokens(self):
        """ Verify that security tokens must match for RPC execution to occur.
        """
        logger = logging.getLogger("rpc.methods")

        # Add an RPC method for testing.
        def security_check():
            logger.info("security check method called.")
            return True

        self.server.add_method(security_check)

        # The 'send_request' method of this test class will always send the
        # server's security token. So define a special function for sending
        # requests via '_send_request' instead.
        def do_check_request(params):
            def request():
                try:
                    self.current_response = send_rpc_request(
                        self.server.url, "security_check", params, 0
                    )
                except Exception as e:
                    self.current_response = e

            return self._send_request(request)

        # Log errors for the 'rpc.methods' logger.
        log_capture = CapturingHandler()
        logger.addHandler(log_capture)
        last_length = [0]

        def next_log_msg():
            records = log_capture.records
            assert len(records) > last_length[0], "no next log message"
            last_length[0] = len(records)
            return records[len(records) - 1].msg

        try:
            # Test with no security token.
            self.assertRaises(RuntimeError, do_check_request, ())
            self.assertIn(next_log_msg(),
                          "client did not send the required "
                          "'security_token' argument")

            # Test with an invalid security token using positional and
            # keyword arguments.
            invalid_token = "NotTheTokenYou'reLookingFor"
            invalid_msg = ("Client sent a security token that did not "
                           "match the server\'s")
            self.assertRaises(RuntimeError, do_check_request,
                              [invalid_token])
            self.assertIn(next_log_msg(), invalid_msg)
            self.assertRaises(RuntimeError, do_check_request,
                              {"security_token": invalid_token})
            self.assertIn(next_log_msg(), invalid_msg)

            # Test with the correct security token.
            valid_msg = "security check method called."
            self.assertTrue(do_check_request([self.server.security_token]))
            self.assertIn(next_log_msg(), valid_msg)
            self.assertTrue(do_check_request({
                "security_token": self.server.security_token}))
            self.assertIn(next_log_msg(), valid_msg)

            # Also check using the normal send_request method that adds the
            # security token automatically.
            self.assertTrue(self.send_request("security_check", []))
            self.assertIn(next_log_msg(), valid_msg)
        finally:
            logger.removeHandler(log_capture)

    def test_list_grammars(self):
        """ Verify that the 'list_grammars' RPC method works correctly. """
        # Load a Grammar with three rules and check that the RPC returns the
        # correct data for them.
        g = Grammar("list_grammars_test")
        g.add_rule(CompoundRule(name="compound", spec="testing",
                                exported=True))
        g.add_rule(MappingRule(name="mapping", mapping={
            "command a": ActionBase(),
            "command b": ActionBase()
        }))
        g.add_rule(Rule(name="base", element=Literal("hello world"),
                        exported=False))
        g.load()

        response = self.send_request("list_grammars", [])
        expected_grammar_data = {
            "name": g.name, "enabled": True, "active": True, "rules": [
                {"name": "compound", "specs": ["testing"],
                 "exported": True, "active": True},
                {"name": "mapping", "specs": ["command a", "command b"],
                 "exported": True, "active": True},
                {"name": "base", "specs": ["hello world"],
                 "exported": False, "active": True}
            ]
        }
        # Check that the loaded grammar appears in the result. It might not
        # be the only grammar and that is acceptable because dragonfly's
        # tests can be run while user grammars are loaded.
        try:
            self.assertIn("result", response)
            self.assertIn(expected_grammar_data, response["result"])
        finally:
            g.unload()

    def test_mimic(self):
        """ Verify that the 'mimic' RPC method works correctly. """
        g = Grammar("mimic_test")
        g.add_rule(CompoundRule(name="compound", spec="testing mimicry",
                                exported=True))
        g.load()

        # Set the grammar as exclusive.
        # The sapi5shared engine apparently requires this for mimic() to
        # work, making the method kind of useless. This does not apply to
        # sapi5inproc.
        g.set_exclusiveness(True)

        response = self.send_request("mimic", ["testing mimicry"])
        try:
            self.assertIn("result", response)
            self.assertEqual(response["result"], True)
        finally:
            g.set_exclusiveness(False)
            g.unload()

    def test_speak(self):
        response = self.send_request("speak", ["testing speak"])
        self.assertIn("result", response)
        self.assertEqual(response["result"], None)

    def test_get_engine_language(self):
        """ Verify that the 'speak' RPC method works correctly. """
        response = self.send_request("get_engine_language", [])
        self.assertIn("result", response)
        self.assertEqual(response["result"], "en")

    def test_recognition_history_methods(self):
        """ Verify that the recognition history RPC methods work correctly.
        """
        # Load a grammar for testing that recognitions are saved.
        g = Grammar("history_test")
        g.add_rule(CompoundRule(name="compound", spec="test history",
                                exported=True))
        g.load()
        g.set_exclusiveness(True)
        try:
            # Test 'register_history()'.
            response = self.send_request("register_history", [])
            self.assertIn("result", response)

            # Test that the method raises an error if used when the observer
            # is already registered.
            self.assertRaises(RuntimeError, self.send_request,
                              "register_history", [])

            # Send a mimic and check that it is returned by the
            # 'get_recognition_history()' method.
            self.send_request("mimic", ["test history"])
            response = self.send_request("get_recognition_history", [])
            self.assertIn("result", response)
            self.assertListEqual(response["result"], [["test", "history"]])

            # Test 'unregister_observer()'.
            response = self.send_request("unregister_history", [])
            self.assertIn("result", response)

            # Test that the method raises an error if used when the observer
            # is not registered.
            self.assertRaises(RuntimeError, self.send_request,
                              "unregister_history", [])
        finally:
            g.set_exclusiveness(False)
            g.unload()


if __name__ == '__main__':
    # Use the "text" engine by default and disable timer manager callbacks
    # to avoid race conditions.
    get_engine("text")._timer_manager.disable()

    from dragonfly.log import setup_log
    setup_log()  # tests require sane logging levels
    unittest.main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Classes for recording audio to recognise.
"""

import threading
import time

import pyaudio


class PyAudioRecorder(object):
    """
    Class for recording audio from a pyaudio input stream.

    This class records on another thread to minimise dropped frames.
    """

    def __init__(self, config, read_interval=0.05):
        self.config = config
        self.read_interval = read_interval
        self._recording = False
        self._thread = None
        self._buffers = []
        self._condition = threading.Condition()
        self._lock = threading.RLock()

    @property
    def recording(self):
        """
        Whether audio is currently being recorded.

        :rtype: bool
        """
        return self._recording

    def start(self):
        """
        Start recording audio in another thread until :meth:`stop` is
        called.

        Audio buffers can be accessed with :meth:`get_buffers`.
        """
        if not self._thread:
            self._recording = True
            self._thread = threading.Thread(target=self._record)
            self._thread.setDaemon(True)
            self._thread.start()

    def get_buffers(self, clear=True):
        """
        Return any stored audio data, optionally clearing the internal list.

        :param clear: whether to clear the internal buffer list.
        :type clear: bool
        :returns: audio buffer list
        :rtype: list
        """
        if self._thread:
            with self._condition:
                self._condition.wait()
                buffers = list(self._buffers)
                if clear:
                    self.clear_buffers()
                return buffers
        else:
            buffers = list(self._buffers)
            if clear:
                self.clear_buffers()
            return buffers

    def clear_buffers(self):
        """
        Clear the internal buffer list.
        """
        with self._lock:
            self._buffers = []

    def stop(self):
        """
        Stop recording audio.
        """
        self._recording = False
        if self._thread:
            self._thread.join(5)
            if not self._thread.is_alive():
                self._thread = None

    def _record(self):
        # Start recording audio on the current thread until stop() is
        # called.
        p = pyaudio.PyAudio()
        channels, rate = self.config.CHANNELS, self.config.RATE
        frames_per_buffer = self.config.FRAMES_PER_BUFFER
        pa_format = pyaudio.get_format_from_width(self.config.SAMPLE_WIDTH)
        stream = p.open(input=True, format=pa_format, channels=channels,
                        rate=rate, frames_per_buffer=frames_per_buffer)

        # Start recognising in a loop
        stream.start_stream()
        while self._recording:
            with self._condition:
                self._buffers.append(stream.read(frames_per_buffer))

                # Notify waiting threads (if any).
                self._condition.notifyAll()

            # This improves the performance; we don't need to process as
            # much audio as the device can read.
            time.sleep(self.read_interval)

        stream.close()
        p.terminate()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

# pylint: disable=E0401
# This file imports MacOS-only symbols.

from AppKit import NSScreen

from .base_monitor import BaseMonitor
from .rectangle import Rectangle


#===========================================================================
# Monitor class for storing info about a single display monitor.

class DarwinMonitor(BaseMonitor):
    """
    The monitor class used on Mac OS (Darwin).

    Monitors for Mac OS are **not** automatically updated because ``AppKit``
    doesn't update monitor information after it is retrieved. A simple
    workaround for this is to restart the Python interpreter after a monitor
    change.

    """

    #-----------------------------------------------------------------------
    # Class methods to create new Monitor objects.

    @classmethod
    def get_all_monitors(cls):
        # Get monitor information from AppKit.
        monitors = []
        for screen in NSScreen.screens():
            # Get the monitor handle.
            handle = screen.deviceDescription()['NSScreenNumber']

            # Create a rectangle object representing the monitor's
            # dimensions and relative position.
            frame = screen.frame()
            dx = frame.size.width - frame.origin.x
            dy = frame.size.height - frame.origin.y
            rectangle = Rectangle(frame.origin.x, frame.origin.y, dx, dy)

            # Get a new or updated monitor object and add it to the list.
            monitors.append(cls.get_monitor(handle, rectangle))

        return monitors

    #-----------------------------------------------------------------------
    # Methods that control attribute access.

    @property
    def name(self):
        """ The name of this monitor. """
        handle = self.handle
        for screen in NSScreen.screens():
            if screen.deviceDescription()['NSScreenNumber'] == handle:
                return screen.localizedName()

        return u"macOS monitor"



"""This file contains the AT-SPI-based accessibility controller implementation
for Linux.

TODO: Reduce duplication between this file and ia2.py.
"""


import logging
import threading
import traceback
import time

from six.moves import queue

from . import base


class Controller(object):
    """Provides access to the AT-SPI subsystem. All accesses to this subsystem
    must be run in a single thread, which is managed here."""

    _log = logging.getLogger("accessibility")

    class Capture(object):
        def __init__(self, closure):
            self.closure = closure
            self.done_event = threading.Event()
            self.exception = None
            self.return_value = None

    def __init__(self):
        self._context = Context()
        # TODO Replace with a completely synchronous queue (size 0).
        self._closure_queue = queue.Queue(1)
        self._focus_queue = []

    def _update_focus(self, event):
        # Add event to a queue for later processing. IAccessible2 functions can
        # cause reentrancy into event handling, so we do this to ensure correct
        # order of processing.
        # TODO: Verify whether this is needed for AT-SPI.
        self._focus_queue.append(event)

    def _process_focus_events(self):
        if not self._focus_queue:
            return

        # Get the latest focus event, skipping obsolete events.
        if len(self._focus_queue) > 1:
            self._log.debug("Skipping %s focus events." % (len(self._focus_queue) - 1))
        event = self._focus_queue[-1]
        self._focus_queue = []

        self._context.focused = Accessible(event.source)
        self._log.debug("Set focused.")

    def _start_blocking(self):
        # Import here so that it can be used in a background thread. The import
        # must be run in the same thread as event registration and handling.
        # TODO: Verify whether this is needed for AT-SPI.
        global pyatspi
        import pyatspi

        # Register event listeners.
        pyatspi.Registry.registerEventListener(self._update_focus,
                                               "object:state-changed:focused")

        thread = threading.Thread(target=pyatspi.Registry.start)
        thread.setDaemon(True)
        thread.start()

        # Perform all AT-SPI operations as they are enqueued.
        # TODO: Process the queues on-demand instead of with a timer.
        while not self.shutdown_event.is_set():
            time.sleep(1e-2)

            # Process events.
            try:
                self._process_focus_events()
            except Exception:
                traceback.print_exc()

            # Process closures.
            while True:
                try:
                    capture = self._closure_queue.get_nowait()
                except queue.Empty:
                    break
                try:
                    capture.return_value = capture.closure(self._context)
                except base.AccessibilityError as exception:
                    capture.exception = exception
                    # Checked exception, don't print.
                    pass
                except Exception as exception:
                    capture.exception = exception
                    # The stack trace won't be captured, so print here.
                    traceback.print_exc()
                capture.done_event.set()

        pyatspi.Registry.stop()

    def start(self):
        self.shutdown_event = threading.Event()
        thread = threading.Thread(target=self._start_blocking)
        thread.setDaemon(True)
        thread.start()

    def stop(self):
        self.shutdown_event.set()

    def run_sync(self, closure):
        capture = self.Capture(closure)
        self._closure_queue.put(capture)
        capture.done_event.wait()
        if capture.exception:
            raise capture.exception
        return capture.return_value


class Context(object):
    """Provides access to the current AT-SPI context, such as focused objects."""

    def __init__(self):
        self.focused = None


class Accessible(object):
    """Wraps an Accessible."""

    def __init__(self, accessible):
        self._accessible = accessible

    def as_text(self):
        try:
            text = self._accessible.queryText()
            return AccessibleTextNode(text)
        except NotImplementedError:
            return None

    def is_editable(self):
        return pyatspi.state.STATE_EDITABLE in self._accessible.getState().getStates()


class BoundingBox(object):
    """Represents a bounding box in screen coordinates."""

    def __init__(self, x, y, width, height):
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def __str__(self):
        return "x=%s, y=%s, width=%s, height=%s" % (self.x, self.y, self.width, self.height)


class AccessibleTextNode(object):
    """Provides a wrapper around a snapshot of AccessibleText. Mutable methods will
    affect the underlying AccessibleText, but the changes will not be reflected
    here."""

    def __init__(self, accessible_text, may_have_cursor=True):
        self.is_leaf = False
        self._text = accessible_text
        self._children = []
        text_length = self._text.characterCount
        text = self._text.getText(0, text_length).decode('utf-8') if text_length > 0 else ""
        cursor_offset = self._text.caretOffset
        if cursor_offset < 0:
            cursor_offset = None
        expanded_text_pieces = []
        child_indices = [i for i, c in enumerate(text) if c == u"\ufffc"]
        if not child_indices:
            self._add_leaf(0, text_length, text, expanded_text_pieces, cursor_offset)
        elif child_indices[0] > 0:
            self._add_leaf(0, child_indices[0], text, expanded_text_pieces, cursor_offset)
        for i, child_index in enumerate(child_indices):
            # TODO Handle case where all embedded objects are non-text and this interface is not supported.
            hypertext = self._text.obj.queryHypertext()
            hyperlink_index = hypertext.getLinkIndex(child_index)
            hyperlink = hypertext.getLink(hyperlink_index)
            # TODO Handle case where embedded object is non-text.
            child = hyperlink.getObject(0).queryText()
            child_node = AccessibleTextNode(child, cursor_offset == child_index)
            self._children.append(child_node)
            expanded_text_pieces.append(child_node.expanded_text)
            end_index = child_indices[i + 1] if i < len(child_indices) - 1 else text_length
            if end_index > child_index + 1:
                self._add_leaf(child_index + 1, end_index, text, expanded_text_pieces, cursor_offset)
        self.expanded_text = "".join(expanded_text_pieces)
        self.cursor = None
        # Only look for cursor if we might have it. This works around a Chrome
        # bug where nodes report the cursor at 0 instead of -1 when they don't
        # have the cursor.
        if may_have_cursor:
            offset = 0
            for child in self._children:
                if child.cursor is not None:
                    self.cursor = offset + child.cursor
                    break;
                offset += len(child.expanded_text)

    def __str__(self):
        return "(" + ", ".join(str(child) for child in self._children) + ")"

    def _add_leaf(self, start, end, text, expanded_text_pieces, cursor_offset):
        child = AccessibleTextLeaf(self._text, text, start, end, cursor_offset)
        self._children.append(child)
        expanded_text_pieces.append(child.expanded_text)

    def set_cursor(self, offset):
        """Sets the cursor to the given offset. Note that the update will not be
        reflected in self.cursor."""

        for child in self._children:
            if offset < len(child.expanded_text):
                child.set_cursor(offset)
                return
            offset -= len(child.expanded_text)

    def get_bounding_box(self, offset):
        for child in self._children:
            if offset < len(child.expanded_text):
                return child.get_bounding_box(offset)
            offset -= len(child.expanded_text)

    def _get_child_and_child_offset(self, offset):
        for child in self._children:
            if offset < len(child.expanded_text):
                return (child, offset)
            offset -= len(child.expanded_text)

    def select_range(self, start, end):
        start_child, start_child_offset = self._get_child_and_child_offset(start)
        end_child, end_child_offset = self._get_child_and_child_offset(end)
        if start_child == end_child and not start_child.is_leaf:
            # Selection is fully contained within a child node.
            start_child.select_range(start_child_offset, end_child_offset)
            return
        if not start_child.is_leaf or not end_child.is_leaf:
            # Selection cannot be made within a single node, which is the only
            # widely supported API.
            raise base.UnsupportedSelectionError()
        # Selection spans one leaf child to another leaf child.
        self._text.setSelection(0,
                                start_child.get_parent_offset(start_child_offset),
                                end_child.get_parent_offset(end_child_offset))


class AccessibleTextLeaf(object):
    """Wrapper around a pure-text segment of an AccessibleText. Mutable methods
    will affect the underlying AccessibleText, but the changes will not be
    reflected here."""

    # Use a broken vertical bar at the end of the text to delimit it from other
    # leaf nodes when expanded. This character was chosen because it is
    # printable and rarely used in practice.
    DELIMITER = u"\u00a6"

    def __init__(self, accessible_text, text, start, end, cursor_offset):
        self.is_leaf = True
        self._text = accessible_text
        self.expanded_text = text[start:end] + self.DELIMITER
        self._start = start
        self._end = end
        self.cursor = cursor_offset - start if cursor_offset is not None and cursor_offset >= start and cursor_offset <= end else None

    def __str__(self):
        return self.expanded_text

    def get_parent_offset(self, offset):
        return self._start + offset

    def set_cursor(self, offset):
        """Sets the cursor to the given offset. Note that the update will not be
        reflected in self.cursor."""

        self._text.setCaretOffset(self.get_parent_offset(offset))

    def get_bounding_box(self, offset):
        screen_relative = 0
        return BoundingBox(*self._text.getCharacterExtents(
            self.get_parent_offset(offset),
            screen_relative))



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
German language implementations of Integer and Digits classes
============================================================================

"""

from ..base.integer_internal  import (MapIntBuilder, CollectionIntBuilder,
                                      MagnitudeIntBuilder, IntegerContentBase)
from ..base.digits_internal   import DigitsContentBase


#---------------------------------------------------------------------------

int_0           = MapIntBuilder({
                                 "null":         0,
                               })
int_1_9_eins    = MapIntBuilder({
                                 "eins":         1,
                                 "zwei":         2,
                                 "drei":         3,
                                 "vier":         4,
                                 "fuenf":        5,
                                 "sechs":        6,
                                 "sieben":       7,
                                 "acht":         8,
                                 "neun":         9,
                               })
int_1_9         = MapIntBuilder({
                                 "ein":          1,
                                 "zwei":         2,
                                 "drei":         3,
                                 "vier":         4,
                                 "fuenf":        5,
                                 "sechs":        6,
                                 "sieben":       7,
                                 "acht":         8,
                                 "neun":         9,
                               })
int_1_eine      = MapIntBuilder({
                                 "eine":         1,
                               })
int_2_9         = MapIntBuilder({
                                 "zwei":         2,
                                 "drei":         3,
                                 "vier":         4,
                                 "fuenf":        5,
                                 "sechs":        6,
                                 "sieben":       7,
                                 "acht":         8,
                                 "neun":         9,
                               })
int_10_19       = MapIntBuilder({
                                 "zehn":        10,
                                 "elf":         11,
                                 "zwoelf":      12,
                                 "dreizehn":    13,
                                 "vierzehn":    14,
                                 "fuenfzehn":   15,
                                 "sechzehn":    16,
                                 "siebzehn":    17,
                                 "achtzehn":    18,
                                 "neunzehn":    19,
                               })
int_20_90_10    = MapIntBuilder({
                                 "zwanzig":      2,
                                 "dreissig":     3,
                                 "vierzig":      4,
                                 "fuenfzig":     5,
                                 "sechzig":      6,
                                 "siebzig":      7,
                                 "achtzig":      8,
                                 "neunzig":      9,
                               })


#---------------------------------------------------------------------------

def int_20_99_func(text):
    # Join any 'und' conjugates with the words on either side.
    return text.replace(" und", "und").replace("und ", "und")


#---------------------------------------------------------------------------

int_20_99       = MagnitudeIntBuilder(
                   factor      = 10,
                   spec        = "[<remainder> und] <multiplier>",
                   multipliers = [int_20_90_10],
                   remainders  = [int_1_9],
                   modifier_function = int_20_99_func,
                   modifier_mode = 1,  # MODE_AUGMENT (default)
                  )
int_and_1_99    = CollectionIntBuilder(
                   spec        = "<element>",
                   set         = [int_1_9_eins, int_10_19, int_20_99],
                  )
int_100s        = MagnitudeIntBuilder(
                   factor      = 100,
                   spec        = "[<multiplier>] hundert [<remainder>]",
                   multipliers = [int_1_9],
                   remainders  = [int_and_1_99],
                  )
int_100big      = MagnitudeIntBuilder(
                   factor      = 100,
                   spec        = "[<multiplier>] hundert [<remainder>]",
                   multipliers = [int_10_19, int_20_99],
                   remainders  = [int_and_1_99]
                  )
int_1000s       = MagnitudeIntBuilder(
                   factor      = 1000,
                   spec        = "[<multiplier>] tausend [<remainder>]",
                   multipliers = [int_1_9, int_10_19, int_20_99, int_100s],
                   remainders  = [int_and_1_99, int_100s]
                  )
int_1000000     = MagnitudeIntBuilder(
                   factor      = 1000000,
                   spec        = "[<multiplier>] million [<remainder>]",
                   multipliers = [int_1_eine],
                   remainders  = [int_and_1_99, int_100s, int_1000s],
                  )
int_1000000s    = MagnitudeIntBuilder(
                   factor      = 1000000,
                   spec        = "<multiplier> millionen [<remainder>]",
                   multipliers = [int_2_9, int_10_19, int_20_99, int_100s,
                                  int_1000s],
                   remainders  = [int_and_1_99, int_100s, int_1000s],
                  )


#---------------------------------------------------------------------------

class IntegerContent(IntegerContentBase):
    builders = [int_0, int_1_9, int_1_9_eins, int_10_19, int_20_99,
                int_100s, int_100big, int_1000s, int_1000000, int_1000000s]

class DigitsContent(DigitsContentBase):
    digits = [("null"), "eins", "zwei", "drei", "vier",
              "fuenf", "sechs", "sieben", "acht", "neun"]



"""
Utilities for training CMU Pocket Sphinx acoustic models.
"""

import contextlib
import os
import time
import wave


def write_training_data(config, frames, hypothesis):
    """
    Write audio frames and the speech hypothesis to new files.

    Audio frames for null hypotheses will not be recorded.

    :raises: IOError | OSError
    """
    if not hypothesis:
        return

    # Check that the directory exists.
    data_dir = config.TRAINING_DATA_DIR
    if not os.path.isdir(data_dir):
        raise IOError("'%s' is not a directory" % data_dir)

    # Use the current time for the file IDs.
    now = time.time()
    base_filename = "%s-%.2f" % (config.TRANSCRIPT_NAME, now)
    wav_filename = base_filename + ".wav"
    txt_filename = base_filename + ".txt"

    # Open and write to a new wave file.
    # Based on the PyAudio record example:
    # https://people.csail.mit.edu/hubert/pyaudio/#record-example
    wav_path = os.path.join(data_dir, wav_filename)
    with contextlib.closing(wave.open(wav_path, "wb")) as wf:
        wf.setnchannels(config.CHANNELS)
        wf.setsampwidth(config.SAMPLE_WIDTH)
        wf.setframerate(config.RATE)
        wf.writeframes(b''.join(frames))

    txt_path = os.path.join(data_dir, txt_filename)
    with open(txt_path, "w") as f:
        f.write(hypothesis)

def write_transcript_files(config, fileids_path, transcriptions_path):
    """
    Write .fileids and .transcriptions using the files in a directory.

    :raises: IOError | OSError
    """
    # Check that the directory exists.
    data_dir = config.TRAINING_DATA_DIR
    transcript_name = config.TRANSCRIPT_NAME
    if not os.path.isdir(data_dir):
        raise IOError("'%s' is not a directory" % data_dir)

    # Get the relevant files from a directory listing.
    def relevant_entry(x):
        return (
            # Must start with "training" by default and end with .wav.
            x.startswith(transcript_name) and x.endswith(".wav") and

            # There must be an accompanying .txt file.
            os.path.isfile(os.path.join(data_dir, "%s.txt" % x[:-4]))
        )

    base_filenames = [f[0:-4] for f in os.listdir(data_dir)
                       if relevant_entry(f)]

    # Sort by time in ascending order.
    base_filenames = sorted(base_filenames)

    # Process each relevant file, adding content to the transcript files
    # where necessary.
    path1, path2 = fileids_path, transcriptions_path
    with open(path1, "w") as f1, open(path2, "w") as f2:
        for name in base_filenames:
            # Get the hypothesis from the text file.
            txt_path = os.path.join(transcript_name, name + ".txt")
            with open(txt_path, "r") as txt_file:
                hypothesis = txt_file.readline().strip()

            # Skip entries with null hypotheses.
            if not hypothesis:
                continue

            # Write data to both files.
            f1.write("%s\n" % name)
            f2.write("<s> %s </s> (%s)\n" % (hypothesis, name))



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Recognition observer class for the SAPI 5 engine
============================================================================

"""

from ..base                import RecObsManagerBase
from ...grammar.grammar    import Grammar
from ...grammar.rule_base  import Rule
from ...grammar.elements   import Impossible


#---------------------------------------------------------------------------

class Sapi5RecObsManager(RecObsManagerBase):

    def __init__(self, engine):
        RecObsManagerBase.__init__(self, engine)
        self._grammar = Sapi5RecObsGrammar(self)

    def _activate(self):
        self._grammar.load()

    def _deactivate(self):
        self._grammar.unload()


#---------------------------------------------------------------------------

class Sapi5RecObsGrammar(Grammar):

    def __init__(self, manager):
        self._manager = manager
        name = "_recobs_grammar"
        Grammar.__init__(self, name, description=None, context=None)

        rule = Rule(element=Impossible(), exported=True)
        self.add_rule(rule)

    #-----------------------------------------------------------------------
    # Callback methods for handling utterances and recognitions.
    # This grammar does not receive recognized words due to a limitation
    # with WSR. The engine instead notifies observers by using the grammar
    # that successfully recognized the words, if any words were recognized.

    def process_begin(self, executable, title, handle):
        self._manager.notify_begin()

    def process_recognition(self, words):
        # Successfully recognising this grammar should be impossible.
        raise RuntimeError("Recognition observer received an unexpected"
                           " recognition: %s" % (words,))

    def process_recognition_failure(self, results):
        self._manager.notify_failure(results)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#


"""
This module initializes the mouse interface for the current platform.
"""

import os
import sys

# Import mouse events common to each platform.
from ._base import (EventBase, PauseEvent, MoveEvent, MoveRelativeEvent,
                    MoveScreenEvent, MoveWindowEvent)


# Import the mouse functions and classes for the current platform.
# Always use the base classes for building documentation.
DOC_BUILD = bool(os.environ.get("SPHINX_BUILD_RUNNING"))
if sys.platform.startswith("win") and not DOC_BUILD:
    from ._win32 import (
        ButtonEvent, get_cursor_position, set_cursor_position,
        PLATFORM_BUTTON_FLAGS, PLATFORM_WHEEL_FLAGS
    )

elif ((os.environ.get("XDG_SESSION_TYPE") == "x11" or
       sys.platform == "darwin") and not DOC_BUILD):
    from ._pynput import (
        ButtonEvent, get_cursor_position, set_cursor_position,
        PLATFORM_BUTTON_FLAGS, PLATFORM_WHEEL_FLAGS
    )

else:
    # No mouse interface is available. Dragonfly can function
    # without this functionality, so don't raise an error or log any
    # messages. Errors/messages will occur later if the mouse is used.
    from ._base import (
        BaseButtonEvent as ButtonEvent, get_cursor_position,
        set_cursor_position, PLATFORM_BUTTON_FLAGS, PLATFORM_WHEEL_FLAGS
    )


# Ensure all button and wheel flag names are accounted for.
# Used for debugging. Uncomment if adding new buttons or aliases.
# _BUTTON_NAMES = ["left", "right", "middle", "four", "five"]
# _WHEEL_NAMES = ["wheelup", "stepup", "wheeldown", "stepdown",
#                 "wheelright", "stepright", "wheelleft", "stepleft"]
# assert sorted(PLATFORM_BUTTON_FLAGS.keys()) == sorted(_BUTTON_NAMES),\
#     "Mouse implementation is missing button flags"
# assert sorted(PLATFORM_WHEEL_FLAGS.keys()) == sorted(_WHEEL_NAMES),\
#     "Mouse implementation is missing wheel flags"



# encoding: utf-8
#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
.. _RefKey:

Key action
============================================================================

This section describes the :class:`Key` action object.  This
type of action is used for sending keystrokes to the foreground
application.  This works on Windows, Mac OS and with X11 (e.g. on Linux).
Examples of how to use this class are given in :ref:`RefKeySpecExamples`.

To use this class on X11/Linux, the
`xdotool <https://www.semicomplete.com/projects/xdotool/>`__ program must be
installed and the ``XDG_SESSION_TYPE`` environment variable set to ``x11``.
This class does **not** support typing keys in Wayland sessions.

.. _RefKeySpec:

Keystroke specification format
............................................................................

The *spec* argument passed to the :class:`Key` constructor specifies which
keystroke events will be emulated.  It is a string consisting of one or
more comma-separated keystroke elements.  Each of these elements has one
of the following two possible formats:

Normal press-release key action, optionally repeated several times:
   [*modifiers* ``-``] *keyname* [``/`` *innerpause*] [``:`` *repeat*] [``/`` *outerpause*]

Press-and-hold a key, or release a held-down key:
   [*modifiers* ``-``] *keyname* ``:`` *direction* [``/`` *outerpause*]

The different parts of the keystroke specification are as follows.  Note
that only *keyname* is required; the other fields are optional.

 - *modifiers* --
   Modifiers for this keystroke.  These keys are held down
   while pressing the main keystroke.
   Can be zero or more of the following:

    - ``a`` -- alt key
    - ``c`` -- control key
    - ``s`` -- shift key
    - ``w`` -- Windows key

 - *keyname* --
   Name of the keystroke.  Valid names are listed in
   :ref:`RefKeySpecNames`.
 - *innerpause* --
   The time to pause between repetitions of this keystroke.
   It should be given in hundredths of a second.  For
   example, "20" will pause for 20/100s = 0.2 seconds.
 - *repeat* --
   The number of times this keystroke should be repeated.
   If not specified, the key will be pressed and released once.
 - *outerpause* --
   The time to pause after this keystroke.
   It should be given in hundredths of a second.  For
   example, "20" will pause for 20/100s = 0.2 seconds.
 - *direction* --
   Whether to press-and-hold or release the key.  Must be
   one of the following:

    - ``down`` -- press and hold the key
    - ``up`` -- release the key

   Note that releasing a key which is *not* being held down does *not*
   cause an error.  It harmlessly does nothing.


.. _RefKeySpecNames:

Key names
............................................................................

 - Lowercase letter keys: ``a`` or ``alpha``, ``b`` or ``bravo``,
   ``c`` or ``charlie``, ``d`` or ``delta``, ``e`` or ``echo``,
   ``f`` or ``foxtrot``, ``g`` or ``golf``, ``h`` or ``hotel``,
   ``i`` or ``india``, ``j`` or ``juliet``, ``k`` or ``kilo``,
   ``l`` or ``lima``, ``m`` or ``mike``, ``n`` or ``november``,
   ``o`` or ``oscar``, ``p`` or ``papa``, ``q`` or ``quebec``,
   ``r`` or ``romeo``, ``s`` or ``sierra``, ``t`` or ``tango``,
   ``u`` or ``uniform``, ``v`` or ``victor``, ``w`` or ``whisky``,
   ``x`` or ``xray``, ``y`` or ``yankee``, ``z`` or ``zulu``
 - Uppercase letter keys: ``A`` or ``Alpha``, ``B`` or ``Bravo``,
   ``C`` or ``Charlie``, ``D`` or ``Delta``, ``E`` or ``Echo``,
   ``F`` or ``Foxtrot``, ``G`` or ``Golf``, ``H`` or ``Hotel``,
   ``I`` or ``India``, ``J`` or ``Juliet``, ``K`` or ``Kilo``,
   ``L`` or ``Lima``, ``M`` or ``Mike``, ``N`` or ``November``,
   ``O`` or ``Oscar``, ``P`` or ``Papa``, ``Q`` or ``Quebec``,
   ``R`` or ``Romeo``, ``S`` or ``Sierra``, ``T`` or ``Tango``,
   ``U`` or ``Uniform``, ``V`` or ``Victor``, ``W`` or ``Whisky``,
   ``X`` or ``Xray``, ``Y`` or ``Yankee``, ``Z`` or ``Zulu``
 - Number keys: ``0`` or ``zero``, ``1`` or ``one``, ``2`` or ``two``,
   ``3`` or ``three``, ``4`` or ``four``, ``5`` or ``five``,
   ``6`` or ``six``, ``7`` or ``seven``, ``8`` or ``eight``,
   ``9`` or ``nine``
 - Symbol keys: ``!`` or ``bang`` or ``exclamation``, ``@`` or ``at``,
   ``#`` or ``hash``, ``$`` or ``dollar``, ``%`` or ``percent``,
   ``^`` or ``caret``, ``&`` or ``and`` or ``ampersand``,
   ``*`` or ``star`` or ``asterisk``,
   ``(`` or ``leftparen`` or ``lparen``,
   ``)`` or ``rightparen`` or ``rparen``, ``minus`` or ``hyphen``,
   ``_`` or ``underscore``, ``+`` or ``plus``, ````` or ``backtick``,
   ``~`` or ``tilde``, ``[`` or ``leftbracket`` or ``lbracket``,
   ``]`` or ``rightbracket`` or ``rbracket``,
   ``{`` or ``leftbrace`` or ``lbrace``,
   ``}`` or ``rightbrace`` or ``rbrace``, ``\\`` or ``backslash``,
   ``|`` or ``bar``, ``colon``, ``;`` or ``semicolon``,
   ``'`` or ``apostrophe`` or ``singlequote`` or ``squote``,
   ``"`` or ``quote`` or ``doublequote`` or ``dquote``, ``comma``,
   ``.`` or ``dot``, ``slash``,
   ``<`` or ``lessthan`` or ``leftangle`` or ``langle``,
   ``>`` or ``greaterthan`` or ``rightangle`` or ``rangle``,
   ``?`` or ``question``, ``=`` or ``equal`` or ``equals``
 - Whitespace and editing keys: ``enter``, ``tab``, ``space``,
   ``backspace``, ``delete`` or ``del``
 - Main modifier keys: ``shift``, ``control`` or ``ctrl``, ``alt``
 - Right modifier keys: ``rshift``, ``rcontrol`` or ``rctrl``, ``ralt``
 - Special keys: ``escape``, ``insert``, ``pause``, ``win``, ``rwin``,
   ``apps`` or ``popup``, ``snapshot`` or ``printscreen``
 - Lock keys: ``scrolllock``, ``numlock``, ``capslock``
 - Navigation keys: ``up``, ``down``, ``left``, ``right``,
   ``pageup`` or ``pgup``, ``pagedown`` or ``pgdown``, ``home``, ``end``
 - Number pad keys: ``npmul``, ``npadd``, ``npsep``, ``npsub``,
   ``npdec``, ``npdiv``, ``numpad0`` or ``np0``, ``numpad1`` or ``np1``,
   ``numpad2`` or ``np2``, ``numpad3`` or ``np3``,
   ``numpad4`` or ``np4``, ``numpad5`` or ``np5``,
   ``numpad6`` or ``np6``, ``numpad7`` or ``np7``,
   ``numpad8`` or ``np8``, ``numpad9`` or ``np9``
 - Function keys: ``f1``, ``f2``, ``f3``, ``f4``, ``f5``, ``f6``,
   ``f7``, ``f8``, ``f9``, ``f10``, ``f11``, ``f12``, ``f13``, ``f14``,
   ``f15``, ``f16``, ``f17``, ``f18``, ``f19``, ``f20``, ``f21``,
   ``f22``, ``f23``, ``f24``
 - Multimedia keys: ``volumeup`` or ``volup``,
   ``volumedown`` or ``voldown``, ``volumemute`` or ``volmute``,
   ``tracknext``, ``trackprev``, ``playpause``, ``browserback``,
   ``browserforward``


.. _RefKeySpecExamples:

Example key actions
............................................................................

The following code types the text "Hello world!" into the foreground
application: ::

    Key("H, e, l, l, o, space, w, o, r, l, d, exclamation").execute()

The following code is a bit more useful, as it saves the current file with
the name "dragonfly.txt" (this works for many English-language
applications): ::

    action = Key("a-f, a/50") + Text("dragonfly.txt") + Key("enter")
    action.execute()

The following code selects the next four lines by holding down the *shift*
key, slowly moving down 4 lines, and then releasing the *shift* key: ::

    Key("shift:down, down/25:4, shift:up").execute()

The following code locks the screen by pressing the *Windows* key together
with the *l* key: ::

    Key("w-l").execute()


Windows key support
............................................................................

Keyboard events sent by :class:`Key` actions on Windows are calculated using
the current foreground window's keyboard layout. The class will fallback on
Unicode events for keys not typeable with the current layout.

The :class:`Key` action can be used to type arbitrary Unicode characters on
Windows using the `relevant Windows API
<https://docs.microsoft.com/en-us/windows/desktop/api/winuser/ns-winuser-tagkeybdinput#remarks>`__.
This is disabled by default because it ignores the up/down status of
modifier keys (e.g. ctrl).

It can be enabled by changing the ``unicode_keyboard`` setting in
`~/.dragonfly2-speech/settings.cfg` to ``True``: ::

    unicode_keyboard = True

The ``use_hardware`` parameter can be set to ``True`` if you need to
selectively require hardware events for a :class:`Key` action: ::

    # Passing use_hardware=True will guarantee that Ctrl+C is always
    # pressed, regardless of the layout. See below.
    Key("c-c", use_hardware=True).execute()

If the Unicode keyboard is not enabled or the ``use_hardware`` parameter is
``True``, then no keys will be typed and an error will be logged for
untypeable keys: ::

   action.exec (ERROR): Execution failed: Keyboard interface cannot type this character: 'μ'

Keys in ranges 0-9, a-z and A-Z are always typeable. If keys in these ranges
cannot be typed using the current keyboard layout, then the equivalent key
will be used instead. For example, the following code will result in the 'я'
key being pressed when using the main Cyrillic keyboard layout: ::

   # This is equivalent to Key(u"я, Я, c-я").
   Key("z, Z, c-z", use_hardware=True).execute()

Unlike the :class:`Text` action, individual :class:`Key` actions can send
both hardware *and* Unicode events. So the following example will work if
the Unicode keyboard is enabled: ::

    # Type 'σμ' and then press Ctrl+Z.
    Key(u"σ, μ, c-z").execute()


X11 key support
............................................................................

The :class:`Key` action can be used to type arbitrary keys and Unicode
characters on X11/Linux. It is not limited to the key names listed above,
although all of them will work too.

Unicode characters are supported on X11 by passing their Unicode code point
to the keyboard implementation. For example, the character ``'€'`` is
converted to ``'U20AC'``. The Unicode code point can also be passed
directly, e.g. with ``Key('U20AC')``.

Unlike on Windows, the :class:`Key` action is able to use modifiers with
Unicode characters on X11.

This class does **not** support typing keys in Wayland sessions.


Example X11 key actions
----------------------------------------------------------------------------

In addition to the examples in the previous section, the following example
will work on X11/Linux.

The following code will type 'σμ' into the foreground application and then
press ctrl+z: ::

    Key("σ,μ,c-z").execute()

The following code will press 'µ' while holding control and alt: ::

    Key("ca-μ").execute()

The following code will press the browser refresh multimedia key: ::

    Key("XF86Refresh").execute()

Although this key is not defined in dragonfly's typeables list, it still
works because it is passed directly to xdotool. X11 (Xorg) multimedia keys
can be found online: `XF86 keyboard symbols
<https://wiki.linuxquestions.org/wiki/XF86_keyboard_symbols>`__.


Key class reference
............................................................................

"""

import collections

from .action_base           import ActionError
from .action_base_keyboard  import BaseKeyboardAction
from .typeables             import typeables

#---------------------------------------------------------------------------


class Key(BaseKeyboardAction):
    """
        Keystroke emulation action.

        Constructor arguments:
         - *spec* (*str*) -- keystroke specification
         - *static* (boolean) -- flag indicating whether the
           specification contains dynamic elements
         - *use_hardware* (boolean) --
           if *True*, send keyboard events using hardware emulation instead
           of as Unicode text. This will respect the up/down status of
           modifier keys.

        The format of the keystroke specification *spec* is described in
        :ref:`RefKeySpec`.

        This class emulates keyboard activity by sending keystrokes to the
        foreground application.  It does this using Dragonfly's keyboard
        interface for the current platform.  The implementation for Windows
        uses the ``sendinput()`` Win32 API function.  The implementation
        for Mac OS uses
        `pynput <https://pynput.readthedocs.io/en/latest/>`__. The
        implementation for X11/Linux uses `xdotool
        <https://www.semicomplete.com/projects/xdotool/>`__.

    """

    # Various keystroke specification format parameters.
    _key_separator = ","
    _delimiter_characters = ":/"
    _modifier_prefix_delimiter = "-"
    _modifier_prefix_characters = {
        'a': typeables["alt"],
        'c': typeables["control"],
        's': typeables["shift"],
        'w': typeables["win"],
        }
    interval_factor = 0.01
    interval_default = 0.0

    # Keystroke event data type.
    EventData = collections.namedtuple(
        "EventData", "keyname "
                     "direction "
                     "modifiers "
                     "inner_pause "
                     "repeat "
                     "outer_pause "
    )

    def _parse_spec(self, spec):
        # Iterate through the keystrokes specified in spec, parsing
        #  each individually.
        events = []
        for single in spec.split(self._key_separator):
            event_data = self._parse_single(single)
            if not event_data:
                continue
            events.append(event_data)
        return events

    def _parse_single(self, spec):
        # pylint: disable=R0912,R0914,R0915
        # Suppress warnings about too many branches, variables and
        # statements.

        # Remove leading and trailing whitespace.
        spec = spec.strip()
        if not spec:
            return None

        # Parse modifier prefix.
        index = spec.find(self._modifier_prefix_delimiter)
        if index != -1:
            s = spec[:index]
            index += 1
            modifiers = []
            for c in s:
                if c not in self._modifier_prefix_characters:
                    raise ActionError("Invalid modifier"
                                      " prefix character: %r" % c)
                m = self._modifier_prefix_characters[c]
                if m in modifiers:
                    raise ActionError("Double modifier"
                                      " prefix character: %r" % c)
                modifiers.append(m)
        else:
            index = 0
            modifiers = ()

        inner_pause = None
        special = None
        outer_pause = None

        delimiters = [(c, i + index) for i, c in enumerate(spec[index:])
                      if c in self._delimiter_characters]
        delimiter_sequence = "".join([d[0] for d in delimiters])
        delimiter_index = [d[1] for d in delimiters]

        if delimiter_sequence == "":
            keyname = spec[index:]
        elif delimiter_sequence == "/":
            keyname = spec[index:delimiter_index[0]]
            outer_pause = spec[delimiter_index[0]+1:]
        elif delimiter_sequence == "/:":
            keyname = spec[index:delimiter_index[0]]
            inner_pause = spec[delimiter_index[0]+1:delimiter_index[1]]
            special = spec[delimiter_index[1]+1:]
        elif delimiter_sequence == "/:/":
            keyname = spec[index:delimiter_index[0]]
            inner_pause = spec[delimiter_index[0]+1:delimiter_index[1]]
            special = spec[delimiter_index[1]+1:delimiter_index[2]]
            outer_pause = spec[delimiter_index[2]+1:]
        elif delimiter_sequence == ":":
            keyname = spec[index:delimiter_index[0]]
            special = spec[delimiter_index[0]+1:]
        elif delimiter_sequence == ":/":
            keyname = spec[index:delimiter_index[0]]
            special = spec[delimiter_index[0]+1:delimiter_index[1]]
            outer_pause = spec[delimiter_index[1]+1:]
        else:
            raise ActionError("Invalid key spec: %s" % spec)

        if inner_pause is not None:
            s = inner_pause
            try:
                inner_pause = float(s) * self.interval_factor
                if inner_pause < 0: raise ValueError
            except ValueError:
                raise ActionError("Invalid inner pause value: %r,"
                                  " should be a positive number." % s)
        if outer_pause is not None:
            s = outer_pause
            try:
                outer_pause = float(s) * self.interval_factor
                if outer_pause < 0: raise ValueError
            except ValueError:
                raise ActionError("Invalid outer pause value: %r,"
                                  " should be a positive number." % s)
        else:
            outer_pause = self.interval_default * self.interval_factor

        direction = None; repeat = 1
        if special is not None:
            if special == "down":   direction = True
            elif special == "up":   direction = False
            else:
                s = special
                try:
                    repeat = int(s)
                    if repeat < 0: raise ValueError
                except ValueError:
                    raise ActionError("Invalid repeat value: %r,"
                                      " should be a positive integer." % s)

        if direction is not None:
            if modifiers:
                raise ActionError("Cannot use direction with modifiers.")
            if inner_pause is not None:
                raise ActionError("Cannot use direction with inner pause.")

        return self.EventData(keyname, direction, modifiers, inner_pause,
                              repeat, outer_pause)

    def _execute_events(self, events):
        # Calculate keyboard events from events (event data).
        use_hardware = self.require_hardware_events()
        keyboard_events = []
        for event_data in events:
            events_single = self._calc_events_single(event_data,
                                                     use_hardware)
            keyboard_events.extend(events_single)

        # Send keyboard events.
        self._keyboard.send_keyboard_events(keyboard_events)
        return True

    def _calc_events_single(self, event_data, use_hardware):
        (keyname, direction, modifiers, inner_pause, repeat,
         outer_pause) = event_data

        # Get a Typeable object for the key, if possible.
        typeable = self._get_typeable(event_data.keyname, use_hardware)

        # Raise an error message if a Typeable could not be retrieved.
        if typeable is None:
            error_message = ("Keyboard interface cannot type this "
                             "character: %r (in %r)" %
                             (keyname, self._spec))
            raise ActionError(error_message)

        # Calculate keyboard events using the Typeable and event data.
        if direction is None:
            if inner_pause is None:
                inner_pause = self.interval_default * self.interval_factor
            if repeat == 0:
                events = []
            else:
                events = []
                for m in modifiers:
                    events.extend(m.on_events())
                for _ in range(repeat - 1):
                    events.extend(typeable.events(inner_pause))
                events.extend(typeable.events(outer_pause))
                for m in modifiers[-1::-1]:
                    events.extend(m.off_events())
        else:
            if direction:
                events = typeable.on_events(outer_pause)
            else:
                events = typeable.off_events(outer_pause)
        return events

    def __str__(self):
        return '[{!r}]'.format(self._spec)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Paste action
============================================================================

"""

import sys

from six import string_types

from ..actions.action_base import ActionError, DynStrActionBase
from ..actions.action_key import Key
from ..windows.clipboard import Clipboard


#---------------------------------------------------------------------------

class Paste(DynStrActionBase):

    """
        Paste-from-clipboard action.

        Constructor arguments:
         - *contents* (*str* | *dict*) -- contents to paste.  This may be a
           simple string to paste, a dynamic action *spec* or a dictionary
           of clipboard format ints to contents (typically strings).
         - *format* (*int*, clipboard format integer) --
           clipboard format.  This argument is ignored if *contents* is a
           dictionary.
         - *paste* (instance derived from *ActionBase*) --
           paste action
         - *static* (boolean) --
           flag indicating whether the
           specification contains dynamic elements

        This action inserts the given *contents* into the system
        clipboard, and then performs the *paste* action to paste it into
        the foreground application.  By default, the *paste* action is the
        :kbd:`Ctrl-v` keystroke or :kbd:`Super-v` on a mac.  The default
        clipboard format used by this action is the *Unicode* text format.

    """

    _default_format = Clipboard.format_unicode

    # Default paste action spec.
    _default_paste_spec = "w-v/20" if sys.platform == "darwin" else "c-v/20"

    # pylint: disable=redefined-builtin
    def __init__(self, contents, format=None, paste=None, static=False):
        if not format:
            format = self._default_format
        if paste is None:
            # Pass use_hardware=True to guarantee that Ctrl+V is always
            # pressed, regardless of the keyboard layout.
            paste = Key(self._default_paste_spec, use_hardware=True)
        if isinstance(contents, string_types):
            spec = contents
            self.contents = None
        else:
            spec = ""
            self.contents = contents
        self.format = format
        self.paste = paste
        DynStrActionBase.__init__(self, spec, static=static)

    def _parse_spec(self, spec):
        if self.contents:
            return self.contents
        else:
            return spec

    def _execute_events(self, events):
        original = Clipboard()
        try:
            original.copy_from_system()
        except Exception as e:
            self._log.warning("Failed to store original clipboard contents:"
                              " %s", e)

        # Store the contents to copy (i.e. *events*) in a Clipboard
        #  instance using the specified (or default) format.  If *events* is
        #  a dictionary, then pass it instead.
        if isinstance(events, dict):
            clipboard = Clipboard(contents=events)
        else:
            clipboard = Clipboard(contents={self.format: events})

        # Copy the contents to the system clipboard and paste using the
        #  paste action.
        clipboard.copy_to_system()
        self.paste.execute()

        # Restore the original clipboard contents afterwards.  This should
        #  clear the clipboard if we failed to store the original contents
        #  above.
        original.copy_to_system()
        return True



# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
RPC server
----------------------------------------------------------------------------

Dragonfly's RPC server handles requests by processing each method through
the current engine's :ref:`multiplexing timer interface <RefEngineTimers>`.
This allows engines to handle requests safely and keeps engine-specific
implementation details out of the *dragonfly.rpc* sub-package.

Security tokens
============================================================================

The RPC server uses mandatory security tokens to authenticate client
requests. This avoids some security issues where, for example, malicious web
pages could send POST requests to the server, even if running on localhost.

If sending requests over an open network, please ensure that the connection
is secure by using TLS or SSH port forwarding.

If the server's :code:`security_token` constructor parameter is not
specified, a new token will be generated and printed to the console. Clients
must specify the security token either as the last positional argument or as
the :code:`security_token` keyword argument.

Errors will be raised if clients send no security token or a token that
doesn't match the server's.

Class reference
============================================================================

"""

import functools
import logging
import threading
import time

from decorator import decorator
from jsonrpc.dispatcher import Dispatcher as BaseDispatcher
from jsonrpc.manager import JSONRPCResponseManager
from werkzeug.wrappers import Request, Response
from werkzeug.serving import run_simple

from .methods import methods
from .security import compare_security_token, generate_security_token
from .util import send_rpc_request
from ..engines import get_engine


class PermissionDeniedError(Exception):
    """
    Error raised if clients send security tokens that don't match the
    server's.
    """


_log = logging.getLogger("rpc.methods")

# --------------------------------------------------------------------------
# RPC method decorator used by the server internally.

@decorator
def _rpc_method(method, *args, **kwargs):
    # Create a thread condition for waiting for the method's result.
    condition = threading.Condition()

    closure = []
    def timer_func():
        try:
            closure.append(method(*args, **kwargs))
        except Exception as e:
            # Log any exceptions.
            _log.exception("Exception occurred during RPC method '%s': %s"
                           % (method.__name__, e))
            closure.append(e)

        # Notify the waiting thread that the result is ready.
        with condition:
            condition.notify()

    # Start a non-repeating timer to execute timer_func().
    get_engine().create_timer(timer_func, 0, repeating=False)

    # Wait for the result.
    with condition:
        condition.wait()

    # Raise an error if it's an exception (json-rpc handles this),
    # otherwise return it.
    result = closure[0]
    if isinstance(result, Exception):
        raise result
    else:
        return result

# --------------------------------------------------------------------------
# JSONRPC Dispatcher sub-class.

class Dispatcher(BaseDispatcher):
    """"""

    def __init__(self, server):
        self.server = server
        super(Dispatcher, self).__init__()

    def add_method(self, f=None, name=None):
        """"""
        if name and not f:
            return functools.partial(self.add_method, name=name)
        elif not name and f:
            name = f.__name__

        if not callable(f):
            raise TypeError("f must be callable")

        def checked_rpc_method(*args, **kwargs):
            assert not (args and kwargs)

            # Pop the security token from args or kwargs. Raise an error if
            # there is no token.
            if args:
                args = list(args)
                rpc_security_token = args.pop(len(args) - 1)
            elif "security_token" in kwargs:
                rpc_security_token = kwargs.pop("security_token")
            else:
                message = ("client did not send the required "
                           "'security_token' argument")
                _log.error(message)
                raise ValueError(message)

            # Check if the received token matches the server's token.
            if not compare_security_token(self.server.security_token,
                                          rpc_security_token):
                message = ("Client sent a security token that did not match"
                           " the server\'s")
                _log.error(message)
                raise PermissionDeniedError(message)

            # The security tokens match, so execute the method.
            return f(*args, **kwargs)

        checked_rpc_method.__name__ = name
        self.method_map[name] = checked_rpc_method
        return f


# --------------------------------------------------------------------------
# RPC server class.

class RPCServer(object):
    """
    RPC server class.

    This class will run a local web server on port 50051 by default. The
    server expects requests using
    `JSON-RPC 2.0 <https://www.jsonrpc.org/specification>`__.

    Constructor arguments:

    - *address* -- address to use (*str*, default: "127.0.0.1")
    - *port* -- port to use (*int*, default: 50051)
    - *ssl_context* -- SSL context object to pass to
      *werkzeug.serving.run_simple* (*SSLContext*, default: None).
    - *threaded* -- whether to use a separate thread to process each request
      (*bool*, default: True).
    - *security_token* -- security token for authenticating clients
      (*str*, default: None). A new token will be generated and printed if
      this parameter is unspecified.

    The *ssl_context* parameter is explained in more detail in Werkzeug's
    `SSL serving documentation
    <http://werkzeug.pocoo.org/docs/0.14/serving/#ssl>`__.

    Secure connections can also be set up using OpenSSH port forwarding with
    a command such as::

        $ ssh -NTf -L 50051:127.0.0.1:50051 <system-with-rpc-server>

    *Minor note*: using an IP address instead of a hostname for the
    *address* parameter should increase performance somewhat, e.g.
    "127.0.0.1" instead of "localhost".

    .. warning::

       Do **not** send requests to the server from the main engine thread;
       thread deadlocks *will* occur if you do this because the main thread
       cannot call timer functions and wait for a response at the same time.
       The RPC framework was designed to be used from *remote* processes.

       Requests will not be processed if the engine is not connected and
       processing speech.


    """

    _log = logging.getLogger("rpc.server")

    def __init__(self, address="127.0.0.1", port=50051, ssl_context=None,
                 threaded=True, security_token=None):
        self._address = address
        self._port = port
        self._ssl_context = ssl_context
        self._threaded = threaded
        if security_token is None:
            security_token = generate_security_token()
            self._log.warning("Generating a new security token because the "
                              "'security_token' parameter was unspecified.")
            print("The generated security token is '%s'." % security_token)

        self.security_token = security_token
        self._thread = None
        self._dispatcher = Dispatcher(self)

        # Add the built-in RPC methods defined in methods.py.
        for method in methods:
            self.add_method(method)

    @Request.application
    def _application(self, request):
        # Add the shutdown method if it is available and has not already
        # been added. This won't be executed via the engine timers.
        shutdown_func = request.environ.get('werkzeug.server.shutdown')
        shutdown_func_name = "shutdown_rpc_server"
        if shutdown_func and shutdown_func_name not in self._dispatcher:
            # Add it to the dispatcher directly.
            self._dispatcher.add_method(shutdown_func, shutdown_func_name)

        # Get a response using the method dispatcher and return it.
        response = JSONRPCResponseManager.handle(
            request.data, self._dispatcher
        )
        return Response(response.json, mimetype='application/json')

    @property
    def url(self):
        """
        The URL to send JSON-RPC requests to.
        """
        # Note: this assumes any non-null SSL context is valid.
        protocol = "https" if self._ssl_context else "http"
        return "%s://%s:%d/jsonrpc" % (protocol, self._address, self._port)

    def send_request(self, method, params, id=0):
        """
        Utility method to send a JSON-RPC request to the server. This will
        block the current thread until a response is received.

        This method is mostly used for testing. If called from the engine's
        main thread, a deadlock *will* occur.

        This will raise an error if the request fails with an error or if
        the server is unreachable.

        The server's security token will automatically be added to the
        ``params`` list/dictionary.

        :param method: name of the RPC method to call.
        :param params: parameters of the RPC method to call.
        :param id: ID of the JSON-RPC request (default: 0).
        :type method: str
        :type params: list | dict
        :type id: int
        :returns: JSON-RPC response
        :rtype: dict
        :raises: RuntimeError
        """
        # Add the server's security token to params.
        security_token = self.security_token
        if isinstance(params, dict):
            params["security_token"] = security_token
        else:
            params.append(security_token)

        return send_rpc_request(self.url, method, params, id)

    def start(self):
        """
        Start the server.

        This method is non-blocking, the RPC server will run on a separate
        daemon thread. This way it is not necessary to call :meth:`stop`
        before the main thread terminates.
        """
        # Handle a previous Thread.join() timeout.
        if self._thread and not self._thread.is_alive():
            self._thread = None

        # Return if the server thread is already running.
        elif self._thread:
            return

        def run():
            # Note: 'threaded' means that each request is handled using a
            # separate thread. This introduces some overhead costs.
            try:
                return run_simple(self._address, self._port,
                                  self._application,
                                  ssl_context=self._ssl_context,
                                  threaded=self._threaded)
            except Exception as e:
                self._log.exception("Exception caught on RPC server thread:"
                                    " %s" % e)

        self._thread = threading.Thread(target=run)
        self._thread.setDaemon(True)
        self._thread.start()

        # Wait a few milliseconds to allow the server to start properly.
        time.sleep(0.1)

    def stop(self):
        """
        Stop the server if it is running.
        """
        # Handle a previous Thread.join() timeout.
        if self._thread and not self._thread.is_alive():
            self._thread = None

        # Return if the server isn't currently running.
        if not self._thread:
            return

        # werkzeug is normally stopped through a request to the server.
        self.send_request("shutdown_rpc_server", [])
        timeout = 5
        self._thread.join(timeout)
        if self._thread.is_alive():
            self._log.warning("RPC server thread failed to stop after %d "
                              "seconds" % timeout)
        else:
            self._thread = None

    def add_method(self, method, name=None):
        """
        Add an RPC method to the server.

        Restarting the server is *not* required for the new method to be
        available.

        This can be used to override method implementations if that is
        desirable.

        This method can also be used as a decorator.

        :param method: the implementation of the RPC method to add.
        :param name: optional name of the RPC method to add. If this is
            None, then :code:`method.__name__` will be used instead.
        :type method: callable
        :type name: str
        """
        if name and not method:
            return functools.partial(self.add_method, name=name)

        new_method = _rpc_method(method)
        self._dispatcher.add_method(new_method, name)

        # Return the original method.
        return method

    def remove_method(self, name):
        """
        Remove an RPC method from the server. This will not raise an error
        if the method does not exist.

        Restarting the server is *not* required for the change to take
        effect.

        :param name: the name of the RPC method to remove.
        :type name: str
        """
        self._dispatcher.pop(name, None)

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



from lark import Lark, Transformer

from dragonfly.grammar.elements_basic import (Literal, Optional, Sequence,
                                              Alternative, Empty)

grammar_string = r"""
?start: alternative

// ? means that the rule will be inlined iff there is a single child
?alternative: sequence ("|" sequence)*
?sequence: single*
         | sequence "{" WORD "}"  -> special

?single: WORD+               -> literal
      | "<" WORD ">"         -> reference
      | "[" alternative "]"  -> optional
      | "(" alternative ")"

// Match anything which is not whitespace or a control character,
// we will let the engine handle invalid words
WORD: /[^\s\[\]<>|(){}]+/

%import common.WS_INLINE
%ignore WS_INLINE
"""

spec_parser = Lark(
    grammar_string,
    parser="lalr"
)

class ParseError(Exception):
    pass

class CompoundTransformer(Transformer):
    """
        Visits each node of the parse tree starting with the leaves
        and working up, replacing lark Tree objects with the
        appropriate dragonfly classes.
    """

    def __init__(self, extras=None, *args, **kwargs):
        self.extras = extras or {}
        Transformer.__init__(self, *args, **kwargs)

    def optional(self, args):
        return Optional(args[0])

    def literal(self, args):
        return Literal(" ".join(args))

    def sequence(self, args):
        return Sequence(args)

    def alternative(self, args):
        return Alternative(args)

    def reference(self, args):
        ref = args[0]
        try:
            return self.extras[ref]
        except KeyError:
            raise Exception("Unknown reference name %r" % (str(ref)))

    def special(self, args):
        child, specifier = args
        if '=' in specifier:
            name, value = specifier.split('=')

            # Try to convert the value to a bool, None or a float.
            if value in ['True', 'False']:
                value = bool(value)
            elif value == 'None':
                value = None
            else:
                try:
                    value = float(value)
                except ValueError:
                    # Conversion failed, value is just a string.
                    pass
        else:
            name, value = specifier, True

        if name in ['weight', 'w']:
            child.weight = float(value)
        elif name in ['test_special']:
            child.test_special = value
        else:
            raise ParseError("Unrecognized special specifier: {%s}" %
                             specifier)

        return child



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Compilers for JSpeech Grammar Format (JSGF) and the CMU Pocket Sphinx speech
recognition engine.
"""

import jsgf
import jsgf.ext
from dragonfly import List, DictList
import dragonfly.grammar.elements as elements_

from ..base import CompilerBase, CompilerError

# noinspection PyUnusedLocal


class JSGFCompiler(CompilerBase):
    """
    Dragonfly compiler for JSpeech Grammar Format (JSGF)

    This class translates dragonfly elements, rules and grammars into
    JSGF expansions, rules and grammars respectively.
    """

    GrammarClass = jsgf.Grammar

    @staticmethod
    def get_reference_name(o):
        # Return a non-nil name string.
        if hasattr(o, "name"):
            if not o.name:
                name = o.__class__.__name__
            else:
                name = o.name
        else:
            # Assume the object is a string.
            name = o

        # JSGF and Pocket Sphinx don't allow spaces in names, but dragonfly
        # does. Work around this by changing any spaces to underscores.
        return name.replace(" ", "_")

    # ----------------------------------------------------------------------
    # Methods for compiling grammars and rules.

    def compile_grammar(self, grammar, *args, **kwargs):
        """
        Take a dragonfly grammar and translate it into a JSGF grammar object
        with methods for compiling the grammar and matching speech.

        :param grammar:
        :param args:
        :param kwargs:
        :return:
        """
        self._log.debug("%s: Compiling grammar %s." % (self, grammar.name))

        # Create a new JSGF Grammar object.
        unknown_words = set()
        result = self.GrammarClass(name=self.get_reference_name(grammar))

        # Compile each dragonfly rule and add it to the new grammar.
        for rule in grammar.rules:
            result.add_rule(self.compile_rule(rule, result, unknown_words))

        # Also compile and add any dragonfly Lists.
        for lst in grammar.lists:
            result.add_rule(self.compile_list(lst, result, unknown_words))

        # Log a warning about unknown words if necessary.
        if unknown_words:
            self._log.warning("Grammar '%s' used words not found in the "
                              "pronunciation dictionary: %s", result.name,
                              ", ".join(sorted(unknown_words)))

        # Return None for empty grammars.
        if not result.rules:
            return

        return result

    def compile_rule(self, rule, *args, **kwargs):
        return jsgf.Rule(
            name=self.get_reference_name(rule),
            visible=rule.exported,
            expansion=self.compile_element(rule.element, *args, **kwargs)
        )

    # ----------------------------------------------------------------------
    # Methods for compiling dragonfly lists and dictionary lists.
    # These have no equivalent in JSGF, so hidden/private rules are used
    # instead.

    def compile_list(self, lst, *args, **kwargs):
        if isinstance(lst, List):
            literal_list = [elements_.Literal(item) for item in lst]
        elif isinstance(lst, DictList):
            keys = list(lst.keys())
            keys.sort()
            literal_list = [elements_.Literal(key) for key in keys]
        else:
            raise CompilerError("Cannot compile dragonfly List %s"
                                % lst)

        return jsgf.HiddenRule(
            self.get_reference_name(lst),
            self.compile_element(elements_.Alternative(literal_list), *args,
                                 **kwargs)
        )

    def recompile_list(self, lst, jsgf_grammar):
        # Used from the GrammarWrapper class to get an updated list and any
        # unknown words.
        unknown_words = set()
        return (self.compile_list(lst, jsgf_grammar, unknown_words),
                unknown_words)

    # ----------------------------------------------------------------------
    # Methods for compiling elements.

    def compile_element(self, element, *args, **kwargs):
        # Look for a compiler method to handle the given element.
        for element_type, compiler in self.element_compilers:
            if isinstance(element, element_type):
                return compiler(self, element, *args, **kwargs)

        # Didn't find a compiler method for this element type.
        raise NotImplementedError("Compiler %s not implemented"
                                  " for element type %s."
                                  % (self, element))

    def _compile_repetition(self, element, *args, **kwargs):
        # Compile the first element only; pyjsgf doesn't support limits on
        # repetition (yet).
        children = element.children
        if len(children) > 1:
            self._log.debug("Ignoring limits of repetition element %s."
                            % element)
        compiled_child = self.compile_element(children[0], *args, **kwargs)
        return jsgf.Repeat(compiled_child)

    def _compile_sequence(self, element, *args, **kwargs):
        # Compile Repetition elements separately.
        if isinstance(element, elements_.Repetition):
            return self._compile_repetition(element, *args, **kwargs)

        children = element.children
        if len(children) > 1:
            return jsgf.Sequence(*[
                self.compile_element(c, *args, **kwargs) for c in children
            ])
        elif len(children) == 1:
            # Skip redundant (1 child) sequences.
            return self.compile_element(children[0], *args, **kwargs)
        else:
            # Compile an Empty element for empty sequences.
            return self.compile_element(elements_.Empty(), *args, **kwargs)

    def _compile_alternative(self, element, *args, **kwargs):
        children = element.children
        if len(children) > 1:
            return jsgf.AlternativeSet(*[
                self.compile_element(c, *args, **kwargs) for c in children
            ])
        elif len(children) == 1:
            # Skip redundant (1 child) alternatives.
            return self.compile_element(children[0], *args, **kwargs)
        else:
            # Compile an Empty element for empty alternatives.
            return self.compile_element(elements_.Empty(), *args, **kwargs)

    def _compile_optional(self, element, *args, **kwargs):
        child = self.compile_element(element.children[0], *args, **kwargs)
        return jsgf.OptionalGrouping(child)

    def _compile_literal(self, element, *args, **kwargs):
        return jsgf.Literal(" ".join(element.words))

    def _compile_rule_ref(self, element, *args, **kwargs):
        name = element.rule.name.replace(" ", "_")
        return jsgf.NamedRuleRef(name)

    def _compile_list_ref(self, element, *args, **kwargs):
        name = element.list.name.replace(" ", "_")
        return jsgf.NamedRuleRef(name)

    def _compile_empty(self, element, *args, **kwargs):
        return jsgf.NullRef()

    def _compile_impossible(self, element, *args, **kwargs):
        return jsgf.VoidRef()

    def _compile_dictation(self, element, *args, **kwargs):
        # JSGF has no equivalent for dictation elements. Instead compile and
        # return an Impossible element that allows dictation to be used,
        # but not matched.
        return self.compile_element(
            elements_.Impossible(), *args, **kwargs
        )


class PatchedRepeat(jsgf.Repeat):
    """
    Repeat class patched to compile JSGF repeats as
    "expansion [expansion]*" to avoid a bug in Pocket Sphinx with the
    repeat operator.
    """
    def compile(self, ignore_tags=False):
        super(PatchedRepeat, self).compile()
        compiled = self.child.compile(ignore_tags)
        if self.tag and not ignore_tags:
            return "(%s)[%s]*%s" % (compiled, compiled, self.tag)
        else:
            return "(%s)[%s]*" % (compiled, compiled)


class SphinxJSGFCompiler(JSGFCompiler):
    """
    JSGF compiler sub-class used by the CMU Pocket Sphinx backend.
    """

    def __init__(self, engine):
        JSGFCompiler.__init__(self)
        self.engine = engine

        # Use a very unlikely phrase to replace unknown words. NullRefs are
        # used instead if words aren't in the vocabulary.
        self.impossible_literal = {
            "en": "impossible " * 20,
        }.get(engine.language, "")

    # ----------------------------------------------------------------------
    # Methods for compiling elements.

    def _compile_repetition(self, element, *args, **kwargs):
        # Compile the first element only; pyjsgf doesn't support limits on
        # repetition (yet).
        children = element.children
        if len(children) > 1:
            self._log.debug("Ignoring limits of repetition element %s."
                            % element)

        # Return a PatchedRepeat instead of a normal Repeat expansion.
        compiled_child = self.compile_element(children[0], *args, **kwargs)
        return PatchedRepeat(compiled_child)

    def _compile_literal(self, element, *args, **kwargs):
        # Build literals as sequences and use <NULL> for unknown words.
        children = []
        for word in element.words:
            if self.engine.check_valid_word(word):
                children.append(jsgf.Literal(word))
            else:
                children.append(self.compile_element(
                    elements_.Impossible(), *args, **kwargs
                ))

                # Save the unknown word.
                args[1].add(word)

        return jsgf.Sequence(*children)

    def _compile_impossible(self, element, *args, **kwargs):
        # Override this to avoid VoidRefs disabling entire rules/grammars.
        # Use a special <_impossible> private rule instead. Only add the
        # special rule if it isn't in the result grammar.
        grammar = args[0]
        if "_impossible" not in grammar.rule_names:
            # Check that the impossible literal contains only valid words.
            words = set(self.impossible_literal.split())
            valid_literal = bool(words)
            for word in words:
                if not valid_literal:
                    break
                if not self.engine.check_valid_word(word):
                    valid_literal = False

            if valid_literal:
                expansion = jsgf.Literal(self.impossible_literal)
            else:
                # Fallback on a NullRef. There are some problems with using
                # these, but they get the job done for simple rules.
                expansion = jsgf.NullRef()
            grammar.add_rule(jsgf.Rule(
                name="_impossible", visible=False, expansion=expansion
            ))

        return jsgf.NamedRuleRef("_impossible")

    # TODO Change this to allow dictation elements to work.
    def _compile_dictation(self, element, *args, **kwargs):
        return self.compile_element(
            elements_.Impossible(), *args, **kwargs
        )



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
This file implements base classes for structured integer grammar
elements.

"""

from ..loader            import language
from ...grammar.elements import (Alternative, Sequence, Optional,
                                 Compound, ListRef, RuleWrap)
from ...grammar.list     import  List


#---------------------------------------------------------------------------
# Base class for integer element classes.

class Integer(Alternative):

    _content = None

    @classmethod
    def _set_content(cls, content):
        """
            Set the Integer class' language-specific contents.

            This is an internal class method.  However, it is also useful
            for testing the Integer framework, as tests for different
            languages can each use this method to force a particular
            Integer language implementation.

        """
        cls._content = content

    def __init__(self, name=None, min=None, max=None, default=None,
                 content=None):
        if content:
            self._content = content
        elif not self._content:
            # Language-specific integer content has not been set yet, so
            #  we set it by retrieving the current speaker language content.
            self._set_content(language.IntegerContent)
        self._builders = self._content.builders

        self._min = min; self._max = max
        children = self._build_children(min, max)
        Alternative.__init__(self, children, name=name, default=default)

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def __repr__(self):
        arguments = []
        if self.name is not None:
            arguments = ["%r" % self.name]
        if self._min is not None or self._max is not None:
            arguments.append("%s" % self._min)
            arguments.append("%s" % self._max)
        return "%s(%s)" % (self.__class__.__name__, ",".join(arguments))

    #-----------------------------------------------------------------------
    # Methods for load-time setup.

    def _build_children(self, min, max):
        memo = {}
        children = [c.build_element(min, max, memo)
                    for c in self._builders]
        return [c for c in children if c]


#---------------------------------------------------------------------------
# Integer reference class.

class IntegerRef(RuleWrap):

    def __init__(self, name, min, max, default=None):
        element = Integer(None, min, max)
        RuleWrap.__init__(self, name, element, default=default)

class ShortIntegerRef(RuleWrap):

    def __init__(self, name, min, max, default=None):
        element = Integer(None, min, max, content=language.ShortIntegerContent)
        RuleWrap.__init__(self, name, element, default=default)

#---------------------------------------------------------------------------



# This file was part of Aenea
#
# Aenea is free software: you can redistribute it and/or modify it under
# the terms of version 3 of the GNU Lesser General Public License as
# published by the Free Software Foundation.
#
# Aenea is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Aenea.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (2014) Alex Roper
# Alex Roper <alex@aroper.net>

# This keyboard implementation is based on Aenea's x11_xdotool.py file.

import logging
import os
import subprocess

from ._x11_base import BaseX11Keyboard, KEY_TRANSLATION


class XdotoolKeyboard(BaseX11Keyboard):
    """Static class for typing keys with xdotool."""

    _log = logging.getLogger("keyboard")

    # xdotool command
    xdotool = "xdotool"

    @classmethod
    def send_keyboard_events(cls, events):
        """
        Send a sequence of keyboard events.

        Positional arguments:
        events -- a sequence of tuples of the form
            (keycode, down, timeout), where
                keycode (str): key symbol.
                down (boolean): True means the key will be pressed down,
                    False means the key will be released.
                timeout (float): number of seconds to sleep after
                    the keyboard event.

        """
        cls._log.debug("Keyboard.send_keyboard_events %r", events)

        # Return early if there are no events (e.g. for Key("")).
        if not events:
            return

        arguments = []
        for event in events:
            (key, down, timeout) = event
            key = KEY_TRANSLATION.get(key, key)

            # Add arguments for pressing keys.
            if down:
                arguments += ['keydown', '--delay', '0', key]
            else:
                arguments += ['keyup', '--delay', '0', key]

            # Instruct xdotool to sleep after the keyboard event if
            # necessary.
            if timeout:
                arguments += ['sleep', '%.3f' % timeout]

        # Press/release the keys using xdotool, catching any errors.
        command = [cls.xdotool] + arguments
        readable_command = ' '.join(command)
        cls._log.debug(readable_command)
        try:
            # Fork the process with setsid() if on a POSIX system such as
            # Linux.
            kwargs = {}
            if os.name == 'posix':
                kwargs.update(dict(preexec_fn=os.setsid))

            # Execute the xdotool child process.
            p = subprocess.Popen(command, **kwargs)
            if p.wait() > 0:
                raise RuntimeError("xdotool command exited with non-zero "
                                   "return code %d" % p.returncode)
        except Exception as e:
            cls._log.exception("Failed to execute xdotool command '%s': "
                               "%s", readable_command, e)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Window mover classes
============================================================================

"""

import math
import random
import threading
import time

from .point        import Point
from .rectangle    import Rectangle, unit


#===========================================================================

def linear_fraction_generator():
    def _linear_fraction_generator(count):
        step = float(1) / count
        for index in range(count):
            yield step * index
    return _linear_fraction_generator

def sine_fraction_generator(begin=-0.5, end=0.5):
    begin_radians = -begin * math.pi
    end_radians   = -end   * math.pi
    def _sine_fraction_generator(count):
        step = (end_radians - begin_radians) / count

        sample_points = [(step*i - begin_radians) for i in range(count)]
        fractions = [math.sin(x) for x in sample_points]
        minimum = min(fractions)
        maximum = max(fractions)
        for f in fractions:
            yield (f - minimum) / (maximum - minimum)
    return _sine_fraction_generator


#---------------------------------------------------------------------------

def linear_path():
    def _linear_path(p1, p2, fractions):
        for fraction in fractions:
            p = p1.interpolate(p2, fraction)
            yield p
    return _linear_path

def spline_path():
    distance_min = 0.5
    distance_max = 1.0
    offset_max = 0.7

    def _spline_path(p1, p2, fractions):
        distance_factor = random.uniform(distance_min, distance_max)
        offset_factor = random.uniform(-offset_max, offset_max)
        delta = p2 - p1
        base = p1.interpolate(p2, distance_factor)
        spline_x = base.x + delta.y * offset_factor
        spline_y = base.y - delta.x * offset_factor
        pm = Point(spline_x, spline_y)

        for fraction in fractions:
            p1m = p1.interpolate(pm, fraction)
            pm2 = pm.interpolate(p2, fraction)
            p = p1m.interpolate(pm2, fraction)
            yield p

    return _spline_path

def linear_resize_path(max_count=10):
    def _linear_resize_path(r1, r2, count):
        if count > max_count:  step = 1.0 / max_count
        else:                  step = 1.0 / count
        for index in range(count):
            if index >= max_count:
                yield r2.dx, r2.dy
            else:
                fraction = step * index
                dx = r1.dx + (r2.dx - r1.dx) * fraction
                dy = r1.dy + (r2.dy - r1.dy) * fraction
                yield dx, dy
    return _linear_resize_path


#===========================================================================

# pylint: disable=R0902,R0913
# Suppress warnings about too many instance attributes and constructor
# arguments.

class TimerThread(threading.Thread):
    def __init__(self, timer_callback, interval):
        threading.Thread.__init__(self)
        self._event = threading.Event()
        self._interval = interval
        self._timer_callback = timer_callback

    def stop(self):
        self._event.set()

    def run(self):
        while not self._event.is_set():
            time.sleep(self._interval)
            self._timer_callback()


class PathBase(object):

    _interval = 0.025

    def __init__(self, window, origin, destination, fraction_generator,
                 position_generator, size_generator, speed=1.0):
        self._window = window
        self._origin = origin
        self._destination = destination
        self._speed = speed
        self._fraction_generator = fraction_generator
        self._position_generator = position_generator
        self._size_generator     = size_generator

        distance = (destination.center - origin.center).magnitude
        count = int(distance / 40 / speed)
        if count < 10:
            count = 10
        self._rectangles = self._rectangle_generator(origin, destination, count)

        self._timer = None

    def start(self):
        if self._timer:
            return

        self._timer = TimerThread(self.timer_callback, self._interval)
        self._timer.start()

    def stop(self):
        if not self._timer:
            return

        self._timer.stop()
        self._timer = None
        self._rectangles = None

    def timer_callback(self):
        if not self._rectangles:
            self.stop()
            self._window.set_position(self._destination)
            return
        try:
            rectangle = next(self._rectangles)
        except StopIteration:
            self._rectangles = None
            self.stop()
            rectangle = self._destination
        self._window.set_position(rectangle)

    def _rectangle_generator(self, r1, r2, count):
        fractions = self._fraction_generator(count)
        positions = self._position_generator(r1.center, r2.center, fractions)
        sizes     = self._size_generator(r1, r2, count)
        for position, size in zip(positions, sizes):
            dx, dy = size
            x = position.x - dx/2
            y = position.y - dy/2
            yield Rectangle(x, y, dx, dy)


#===========================================================================

class WindowMover(object):

    def __init__(self, fraction_generator, position_generator,
                 size_generator, speed=1.0):
        self._fraction_generator = fraction_generator
        self._position_generator = position_generator
        self._size_generator     = size_generator
        self._speed = speed

    def move_window(self, window, origin, destination):
        path = PathBase(window, origin, destination,
                        self._fraction_generator, self._position_generator,
                        self._size_generator)
        path.start()


#===========================================================================

window_movers = {
    "spline":   WindowMover(
        sine_fraction_generator(),
        spline_path(),
        linear_resize_path(20),
    ),
    "linear":   WindowMover(
        linear_fraction_generator(),
        linear_path(),
        linear_resize_path(20),
    ),
}



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

# pylint: disable=W0622
# Suppress warnings about redefining the built-in 'id' function.

from .base_window import BaseWindow
from .rectangle import Rectangle


class FakeWindow(BaseWindow):
    """ Fake Window class used when no implementation is available. """

    fake_title = ''
    fake_classname = ''
    fake_executable = ''
    fake_pid = 0

    @classmethod
    def get_foreground(cls):
        return FakeWindow(id=0)

    def __init__(self, id):
        super(FakeWindow, self).__init__(id)
        self._rectangle = Rectangle()

    @classmethod
    def get_all_windows(cls):
        return []

    #-----------------------------------------------------------------------
    # Methods and properties for window attributes.

    def _get_window_text(self):
        return self.fake_title

    def _get_class_name(self):
        return self.fake_classname

    def _get_window_module(self):
        return self.fake_executable

    def _get_window_pid(self):
        return self.fake_pid

    @property
    def is_minimized(self):
        return False

    @property
    def is_maximized(self):
        return False

    @property
    def is_visible(self):
        return True

    #-----------------------------------------------------------------------
    # Methods related to window geometry.

    def get_position(self):
        return self._rectangle

    def set_position(self, rectangle):
        assert isinstance(rectangle, Rectangle)
        self._rectangle = rectangle

    #-----------------------------------------------------------------------
    # Methods for miscellaneous window control.

    def minimize(self):
        pass

    def maximize(self):
        pass

    def restore(self):
        pass

    def close(self):
        pass

    def set_foreground(self):
        pass

    def set_focus(self):
        pass



# coding=utf-8
#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#


import unittest
import string

from dragonfly import parser
from dragonfly import Compound

#===========================================================================


class TestParsers(unittest.TestCase):

    def setUp(self):
        pass

    def test_character_series(self):
        """ Test CharacterSeries parser classes. """

        # Test with ascii characters
        self._test_multiple(
            parser.CharacterSeries(string.ascii_letters),
            [
                ("abc", ["abc"]),
            ],
            must_finish=False)

        # Test the Letters and Alphanumerics classes with a few inputs and
        # outputs
        input_outputs = [
            # Unicode strings
            (u"abc", [u"abc"]),
            # Mix of non-ascii characters
            (u"éèàâêùôöëäïüû", [u"éèàâêùôöëäïüû"]),
            (u"touché", [u"touché"]),
        ]
        self._test_multiple(
            parser.Letters(),
            input_outputs + [(string.digits, [])],
            must_finish=False
        )
        self._test_multiple(
            parser.Alphanumerics(),
            input_outputs + [(string.digits, [string.digits])],
            must_finish=False
        )

    def test_other_alphabets(self):
        """ Test that other alphabets also work. """
        # While the DNS and WSR engines are mostly limited to western
        # European languages (i.e. windows-1252), that doesn't necessarily
        # mean the parser needs to be.
        input_outputs = [
            (u"йцукенгшщзхъфывапролджэячсмитьбю",
             [u"йцукенгшщзхъфывапролджэячсмитьбю"]),
        ]
        self._test_multiple(
            parser.Letters(), input_outputs, must_finish=False
        )
        self._test_multiple(
            parser.Alphanumerics(), input_outputs, must_finish=False
        )

    def test_compound_element(self):
        """ Test the dragonfly Compound element. """
        # Compound is the element that the base dragonfly rule classes use.
        # It uses custom parser elements, so it needs to be tested separately.
        self.assertEqual(
            Compound(spec=u"touché").gstring(),
            u"(touché)"
        )
        self.assertEqual(
            Compound(spec=u"йцукенгшщзхъфывапролджэячсмитьбю").gstring(),
            u"(йцукенгшщзхъфывапролджэячсмитьбю)"
        )

    def test_repetition(self):
        """ Test repetition parser class. """
        word = parser.Letters()
        whitespace = parser.Whitespace()
        p = parser.Repetition(parser.Alternative((word, whitespace)))

        # Test with ascii letters
        input_output = (
            ("abc", ["abc"]),
            ("abc abc", ["abc", " ", "abc"]),
            ("abc abc\t\t\n   cba", ["abc", " ", "abc", "\t\t\n   ",
                                     "cba"]),
            )
        self._test_single(p, input_output)

        # Test with non-ascii letters
        input_output = (
            (u"êùö", [u"êùö"]),
            (u"êùö êùö", [u"êùö", u" ", u"êùö"]),
            (u"êùö êùö\t\t\n   öùê", [
                u"êùö", " ", u"êùö", u"\t\t\n   ", u"öùê"]),
        )
        self._test_single(p, input_output)

    def test_optional_greedy(self):
        """ Test greedy setting of optional parser class. """
        input = "abc"
        p = parser.Sequence([
            parser.Sequence([
                parser.String("a"),
                parser.Optional(parser.String("b"), greedy=False)
                ]),
            parser.Sequence([
                parser.Optional(parser.String("b")),
                parser.String("c")
                ]),
            ])
        expected_output_1 = [["a", None], ["b", "c"]]
        expected_output_2 = [["a", "b"], [None, "c"]]

        state = parser.State(input)
        generator = p.parse(state)
        next(generator)
        root = state.build_parse_tree()
        self.assertEqual(root.value(), expected_output_1)
        next(generator)
        root = state.build_parse_tree()
        self.assertEqual(root.value(), expected_output_2)

    def _test_single(self, parser_element, input_output):
        p = parser.Parser(parser_element)
        for input, output in input_output:
            result = p.parse(input)
            self.assertEqual(result, output)

    def _test_multiple(self, parser_element, input_outputs, must_finish=True):
        p = parser.Parser(parser_element)
        for input, outputs in input_outputs:
            results = p.parse_multiple(input, must_finish)
            self.assertEqual(len(results), len(outputs))
            for index, result, output in zip(list(range(len(results))), results, outputs):
                if isinstance(result, list): result = tuple(result)
                if isinstance(output, list): output = tuple(output)
                self.assertEqual(result, output)


#===========================================================================

if __name__ == "__main__":
    unittest.main()



"""
Dragonfly command module designed for use with the CMU Pocket Sphinx engine,
either with the module loader or as a script.

Shows use of dragonfly Functions, Lists, Repetition, Dictation, IntegerRefs,
Mimic, etc.
"""

from dragonfly import (Dictation, Function, Grammar, IntegerRef, List,
                       ListRef, MappingRule, RecognitionObserver, Mimic,
                       Repetition, get_engine, Text)


if __name__ == '__main__':
    # If this file is being run, not imported, then use the sphinx engine.
    sphinx_engine = get_engine("sphinx")


def print_(x):
    print(x)


# Set up a dragonfly list.
lst = List("lst")

def update_list():
    if lst:
        print("Removing item from list.")
        lst.pop()
    else:
        item = "list item"
        print("Added '%s' to the list." % item)
        lst.append(item)


def write_transcripts():
    engine = get_engine()
    if engine.name == "sphinx":
        engine.write_transcript_files(
            "training.fileids", "training.transcription"
        )


# Define a function for typing multiple numbers.
def type_numbers(numbers):
    Text("".join(map(str, numbers))).execute()


class ExampleRule(MappingRule):
    mapping = {
        # Recognise 'hello' followed by arbitrary dictation.
        # This mapping cannot be matched using Pocket Sphinx because the
        # engine's support for dictation has been temporarily disabled.
        "hello <dictation>": Function(print_, dict(dictation="x")),

        # You still can use Mimic or engine.mimic() to match dictation.
        "say hello world": Mimic("hello", "WORLD"),

        # Update and recognise from a dragonfly list.
        "update list": Function(update_list),
        "<lst>": Function(print_, dict(lst="x")),

        # Command to type numbers, e.g. 'type one two three'.
        "type <numbers>": Function(type_numbers),

        # Write transcript files used for training models.
        "(make|write) transcripts": Function(write_transcripts),
    }

    extras = [
        Dictation("dictation"),
        ListRef("lst", lst),
        Repetition(IntegerRef("n", 1, 20), min=1, max=16,
                   name="numbers"),
    ]


grammar = Grammar("Sphinx engine example")
grammar.add_rule(ExampleRule())
grammar.load()


if __name__ == '__main__':
    # If this file is being run, not imported, set up a recognition observer
    # to print some info.

    class Observer(RecognitionObserver):
        def on_begin(self):
            print("Speech started.")

        def on_recognition(self, words):
            print(" ".join(words))

        def on_failure(self):
            print("Sorry, what was that?")

    observer = Observer()
    observer.register()

    # Start engine's main recognition loop
    try:
        sphinx_engine.recognise_forever()
    except KeyboardInterrupt:
        pass



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This file implements the compiler class for Natlink.
"""


#---------------------------------------------------------------------------

import struct
from six import string_types, text_type
from locale import getpreferredencoding

from ..base import CompilerBase, CompilerError

import dragonfly.grammar.elements as elements_


#===========================================================================

class NatlinkCompiler(CompilerBase):

    #-----------------------------------------------------------------------
    # Methods for compiling grammars.

    def compile_grammar(self, grammar):
        self._log.debug("%s: Compiling grammar %s." % (self, grammar.name))

        compiler = _Compiler()
        for rule in grammar.rules:
            self._compile_rule(rule, compiler)

        compiled_grammar = compiler.compile()
        rule_names = compiler.rule_names
#        print compiler.debug_state_string()
        return (compiled_grammar, rule_names)

    def _compile_rule(self, rule, compiler):
        self._log.debug("%s: Compiling rule %s." % (self, rule.name))
        if rule.imported:
            return
        compiler.start_rule_definition(rule.name, exported=rule.exported)
        self.compile_element(rule.element, compiler)
        compiler.end_rule_definition()

    #-----------------------------------------------------------------------
    # Methods for compiling elements.

    def _compile_sequence(self, element, compiler):
        children = element.children
        if len(children) > 1:
            # Compile Sequence and Repetition elements differently.
            is_rep = isinstance(element, elements_.Repetition)
            if is_rep and element.optimize:
                compiler.start_repetition()
                self.compile_element(children[0], compiler)
                compiler.end_repetition()
            else:
                compiler.start_sequence()
                for c in children:
                    self.compile_element(c, compiler)
                compiler.end_sequence()
        elif len(children) == 1:
            self.compile_element(children[0], compiler)

    def _compile_alternative(self, element, compiler):
        children = element.children
        if len(children) > 1:
            compiler.start_alternative()
            for c in children:
                self.compile_element(c, compiler)
            compiler.end_alternative()
        elif len(children) == 1:
            self.compile_element(children[0], compiler)

    def _compile_optional(self, element, compiler):
        compiler.start_optional()
        self.compile_element(element.children[0], compiler)
        compiler.end_optional()

    def _compile_literal(self, element, compiler):
        # Use the element's words extension property to enable use of
        # quoted words. DNS will receive quoted words as single strings.
        # This can help with accuracy in some cases.
        words = element.words_ext
        if len(words) == 1:
            compiler.add_word(words[0])
        elif len(words) > 1:
            compiler.start_sequence()
            for w in words:
                compiler.add_word(w)
            compiler.end_sequence()

    def _compile_rule_ref(self, element, compiler):
        compiler.add_rule(element.rule.name, imported=element.rule.imported)

    def _compile_list_ref(self, element, compiler):
        compiler.add_list(element.list.name)

    def _compile_dictation(self, element, compiler):
        compiler.add_rule("dgndictation", imported=True)

    def _compile_impossible(self, element, compiler):
        compiler.add_list("_empty_list")

    def _compile_empty(self, element, compiler):
        pass


#===========================================================================
# Internal compiler class which takes care of the binary format
#  used to specify grammars with Dragon NaturallySpeaking.

class _Compiler(object):

    # Numeric values used in the binary form of rule definitions.
    _start_type = 1; _end_type = 2
    _word_type = 3; _rule_type = 4; _list_type = 6
    _seq_value = 1; _alt_value = 2; _rep_value = 3; _opt_value = 4

    def __init__(self):
        self._words = []; self._lists = []; self._rules = []
        self._import_rules = []
        self._export_rules = []
        self._rule_definitions = {}

        self._current_rule_name = None
        self._current_rule_export = None
        self._current_rule_definition = None

    #-----------------------------------------------------------------------

    def start_rule_definition(self, name, exported=False):
        """start defining a rule."""

        # Make sure no rule is being defined at the moment.
        if self._current_rule_name:
            raise CompilerError("Cannot start defining a rule while" \
                               "a different rule is already being defined.")

        assert isinstance(name, string_types), ("The rule name must be a"
                                                " string, received %r." % name)
        self._current_rule_name = name
        self._current_rule_export = exported
        self._current_rule_definition = []

    def end_rule_definition(self):
        """End defining a rule."""

        # Make sure a rule is being defined at the moment.
        if not self._current_rule_name:
            raise CompilerError("Cannot end defining a rule when" \
                               " no rule is being defined.")

        # Make sure that no other rule has been defined with this name.
        if self._current_rule_name in self._rule_definitions:
            raise CompilerError("Rule '%s' defined more than once." % \
                                self._current_rule_name)

        # If this rule has not been used before, register it.
        if self._current_rule_name not in self._rules:
            self._rules.append(self._current_rule_name)
        if self._current_rule_export:
            self._export_rules.append(self._current_rule_name)
        self._rule_definitions[self._current_rule_name] = \
            self._current_rule_definition

        self._current_rule_name = None
        self._current_rule_export = None
        self._current_rule_definition = None

    #-----------------------------------------------------------------------
    # Compound structures in a rule definition.

    def start_sequence(self):
        """start a sequence structure in the rule currently being defined."""
        # Make sure a rule is being defined at the moment.
        if not self._current_rule_name:
            raise CompilerError("Cannot start a sequence because" \
                               " no rule is currently being defined.")
        # Append the start-tag.
        self._current_rule_definition.append(
            (self._start_type, self._seq_value) )

    def end_sequence(self):
        """End a sequence structure in the rule currently being defined."""
        # Make sure a rule is being defined at the moment.
        if not self._current_rule_name:
            raise CompilerError("Cannot end a sequence because" \
                               " no rule is currently being defined.")
        # Append the end-tag.
        self._current_rule_definition.append(
            (self._end_type, self._seq_value) )

    def start_alternative(self):
        """start an alternative structure in the rule currently being defined."""
        # Make sure a rule is being defined at the moment.
        if not self._current_rule_name:
            raise CompilerError("Cannot start an alternative because" \
                               " no rule is currently being defined.")
        # Append the start-tag.
        self._current_rule_definition.append(
            (self._start_type, self._alt_value) )

    def end_alternative(self):
        """End an alternative structure in the rule currently being defined."""
        # Make sure a rule is being defined at the moment.
        if not self._current_rule_name:
            raise CompilerError("Cannot end an alternative because" \
                               " no rule is currently being defined.")
        # Append the end-tag.
        self._current_rule_definition.append(
            (self._end_type, self._alt_value) )

    def start_repetition(self):
        """start a repetition structure in the rule currently being defined."""
        # Make sure a rule is being defined at the moment.
        if not self._current_rule_name:
            raise CompilerError("Cannot start a repetition because" \
                               " no rule is currently being defined.")
        # Append the start-tag.
        self._current_rule_definition.append(
            (self._start_type, self._rep_value) )

    def end_repetition(self):
        """End a repetition structure in the rule currently being defined."""
        # Make sure a rule is being defined at the moment.
        if not self._current_rule_name:
            raise CompilerError("Cannot end a repetition because" \
                               " no rule is currently being defined.")
        # Append the end-tag.
        self._current_rule_definition.append(
            (self._end_type, self._rep_value) )

    def start_optional(self):
        """start a optional structure in the rule currently being defined."""
        # Make sure a rule is being defined at the moment.
        if not self._current_rule_name:
            raise CompilerError("Cannot start a optional because" \
                               " no rule is currently being defined.")
        # Append the start-tag.
        self._current_rule_definition.append(
            (self._start_type, self._opt_value) )

    def end_optional(self):
        """End a optional structure in the rule currently being defined."""
        # Make sure a rule is being defined at the moment.
        if not self._current_rule_name:
            raise CompilerError("Cannot end a optional because" \
                               " no rule is currently being defined.")
        # Append the end-tag.
        self._current_rule_definition.append(
            (self._end_type, self._opt_value) )

    #-----------------------------------------------------------------------
    # Terminal elements in a rule definition.

    def add_word(self, word):
        """Append a literal word to the rule currently being defined."""
        # Make sure a rule is being defined at the moment.
        if not self._current_rule_name:
            raise CompilerError("Cannot add word '%s' because" \
                               " no rule is currently being defined." % word)
        # Determine this word's ID.  If this word has not been used before,
        # register it.
        if word not in self._words:
            self._words.append(word)
            id = len(self._words)
        else:
            id = self._words.index(word) + 1
        # Append the word to the rule currently being defined.
        self._current_rule_definition.append( (self._word_type, id) )

    def add_list(self, list):
        """Append a list to the rule currently being defined."""
        # Make sure a rule is being defined at the moment.
        if not self._current_rule_name:
            raise CompilerError("Cannot add list '%s' because" \
                               " no rule is currently being defined." % list)
        # Determine this list's ID.  If this list has not been used before,
        # register it.
        if list not in self._lists:
            self._lists.append(list)
            id = len(self._lists)
        else:
            id = self._lists.index(list) + 1
        # Append the list to the rule currently being defined.
        self._current_rule_definition.append( (self._list_type, id) )

    def add_rule(self, rule, imported = False):
        """Append a rule reference to the rule currently being defined."""
        # Make sure a rule is being defined at the moment.
        if not self._current_rule_name:
            raise CompilerError("Cannot add rule '%s' because" \
                               " no rule is currently being defined." % rule)
        # Determine this rule's ID.  If this rule has not been used before,
        # register it.
        if rule not in self._rules:
            self._rules.append(rule)
            if imported: self._import_rules.append(rule)
            id = len(self._rules)
        # If this rule has been referenced multiple times, make sure that
        # it has either always being imported or never been imported.
        elif imported != (rule in self._import_rules):
            raise CompilerError("Rule '%s' cannot be referenced as both" \
                               " imported and not imported within a" \
                               " grammar." % rule)
        else:
            id = self._rules.index(rule) + 1
        # Append the rule to the rule currently being defined.
        self._current_rule_definition.append( (self._rule_type, id) )

    #-----------------------------------------------------------------------

    def compile(self):
        """Compile a binary grammar of this compiler's current state."""

        # Make sure no rule is being defined at the moment.
        if self._current_rule_name:
            raise CompilerError("Cannot compile grammar while a rule" \
                               " is being defined.")

        # Grammar header:
        #   - dwType; use 0.
        #   - dwFlags; use 0.
        output = [struct.pack("LL", 0, 0)]

        # Lists of the names and IDs of exports, imports, lists, and words.
        output.append(self._compile_id_chunk(4, self._export_rules, self._rules))
        output.append(self._compile_id_chunk(5, self._import_rules, self._rules))
        output.append(self._compile_id_chunk(6, self._lists, self._lists))
        output.append(self._compile_id_chunk(2, self._words, self._words))

        # List of rule definitions.
        output.append(self._compile_rule_chunk(3))

        # Return a concatenation of the header and chunks.
        return b"".join(output)

    def _compile_id_chunk(self, chunk_id, subset, ordered_superset):
        # Loop through the elements of the superset, and if also present in
        #  the subset create a data entry of its ID and name.  The IDs start
        #  counting at 1.
        elements = []
        for name, id in zip(ordered_superset,
                            range(1, len(ordered_superset) + 1)):

            # Skip names not included in the subset.
            if name not in subset: continue

            # Chunk element:
            #  - dwSize; size of this element in bytes including this header.
            #  - dwNum; the element's ID.
            #  - szName; the element's name terminated by at least one 0.
            # The element's name must be followed by one or more 0
            #  characters, so that its size in bytes is a multiple of 4.
            if isinstance(name, text_type):
                name = name.encode(getpreferredencoding())

            padded_len = (len(name) + 4) & (~3)
            element = struct.pack(("LL%ds" % padded_len),
                padded_len + 8, id, name)
            elements.append(element)

        # Concatenate all the elements.
        element_data = b"".join(elements)

        # Chunk header:
        #  - dwChunkId; words:2, rule definitions:3,
        #     exports:4, imports:5, lists:6
        #  - dwChunkSize; size of this chunk in bytes excluding this header.
        header = struct.pack("LL", chunk_id, len(element_data))

        # Return the header and the elements.
        return header + element_data

    def _compile_rule_chunk(self, chunk_id):
        # Loop through all known rule names, and if they are defined within
        #  this grammar create a rule definition entry.
        definitions = []
        for name, id in zip(self._rules,
                            range(1, len(self._rules) + 1)):

            # Skip imported rules.
            if name in self._import_rules:
                if name in self._rule_definitions:
                    raise CompilerError("Rule '%s' cannot be both imported" \
                                        " and defined in a grammar" % name)
                continue

            # Make sure that a definition has been given.
            if name not in self._rule_definitions:
                raise CompilerError("Rule '%s' is neither imported" \
                                    " nor defined" % name)

            # Build the definition sequence for this rule.
            elements = []
            for t, v in self._rule_definitions[name]:
                # Definition element:
                #  - wType; start:1, end:2, word:3, rule:4, list:6
                #  - wProb; probability rating, use 0.
                #  - dwValue; depends on wType as follows:
                #     - if wType is start or end, then dwValue is one of
                #        sequence:1, alternative:2, repetition:3, optional:4
                #     - if wType is word, rule, or list, then dwValue is
                #        the ID of the corresponding element.
                element = struct.pack("HHL", t, 0, v)
                elements.append(element)

            # Definition header:
            #  - dwSize; size of this rule definition in bytes including
            #     this header.
            #  - dwNum; the ID of this rule.
            definition_size = 8 + sum([len(s) for s in elements])
            definition = struct.pack("LL", definition_size, id)
            definition += b"".join(elements)

            definitions.append(definition)

        # Concatenate all the rule definitions.
        definition_data = b"".join(definitions)

        # Rule definition chunk header:
        #  - dwChunkId; rule definitions:3
        #  - dwChunkSize; size of this chunk in bytes excluding this header.
        header = struct.pack("LL", chunk_id, len(definition_data))

        # Return the header and the rule definitions.
        return header + definition_data

    def _get_rule_names(self):
        return tuple([None] + self._rules)

    rule_names = property(_get_rule_names,
                doc="Read-only access to the list of rule names.")

    #-----------------------------------------------------------------------

    def debug_state_string(self):
        """Debug."""

        import textwrap
        wrapper = textwrap.TextWrapper(subsequent_indent = "   ")
        output = []

        wrapper.initial_indent = "exported rules: "
        output.append(wrapper.wrap(", ".join(self._export_rules)))
        wrapper.initial_indent = "imported rules: "
        output.append(wrapper.wrap(", ".join(self._import_rules)))
        wrapper.initial_indent = "lists: "
        output.append(wrapper.wrap(", ".join(self._lists)))
        wrapper.initial_indent = "words: "
        output.append(wrapper.wrap(", ".join(self._words)))
        wrapper.initial_indent = "rule definitions: "
        output.append(wrapper.wrap(str(self._rule_definitions)))

        return "\n".join(["\n".join(lines) for lines in output if lines])



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This module demonstrates the use of the Dragonfly speech recognition
    library.

"""

from dragonfly import (Grammar, CompoundRule, Dictation)


#---------------------------------------------------------------------------
# Create a compound rule which demonstrates CompoundRule and Choice types.

class AppleRule(CompoundRule):
    spec = "I like apples [<text>]"
    extras = [Dictation("text")]
    def _process_recognition(self, node, extras):
        print("I like apples!  (%s)" % extras.get("text", ""))
        apple_rule.disable()
        banana_rule.enable()

class BananaRule(CompoundRule):
    spec = "I like bananas [<text>]"
    extras = [Dictation("text")]
    def _process_recognition(self, node, extras):
        print("I like bananas!  (%s)" % extras.get("text", ""))
        banana_rule.disable()
        apple_rule.enable()

apple_rule = AppleRule()
banana_rule = BananaRule()


#---------------------------------------------------------------------------
# Create this module's grammar.

grammar = Grammar("fruit toggle")
grammar.add_rule(apple_rule)
grammar.add_rule(banana_rule)


#---------------------------------------------------------------------------
# Load the grammar instance and define how to unload it.

grammar.load()

# Unload function which will be called by natlink at unload time.
def unload():
    global grammar
    if grammar: grammar.unload()
    grammar = None



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
SR back-end package for WSR and SAPI 5
============================================================================

"""

import logging
import platform

_log = logging.getLogger("engine.sapi5")


#---------------------------------------------------------------------------

# Module level singleton instance of this engine implementation.
_engine = None


def is_engine_available(name, **kwargs):
    """
        Check whether SAPI is available.

        :param name: optional human-readable name of the engine to return.
        :type name: str
        :param \\**kwargs: optional keyword arguments passed through to the
            engine for engine-specific configuration.
    """
    global _engine
    if _engine:
        return True

    platform_name = platform.system()
    if platform_name != 'Windows':
        _log.warning("%s is not supported by the SAPI5 engine backend",
                     platform_name)
        return False

    # Attempt to import win32 package required for COM.
    try:
        from win32com.client import Dispatch
    except ImportError as e:
        _log.warning("Failed to import from win32com package: %s. Is it "
                     "installed?", e)
        return False

    try:
        from pywintypes import com_error
    except ImportError as e:
        _log.warning("Failed to import from the pywintypes package: %s. Is "
                     "it installed?", e)
        return False

    # Attempt to connect to SAPI using the correct dispatch name.
    from .engine import Sapi5Engine, Sapi5InProcEngine

    # Use the in-process engine by default, otherwise use the shared engine.
    if not name or name == "sapi5inproc":
        dispatch_name = Sapi5InProcEngine.recognizer_dispatch_name
    else:
        dispatch_name = Sapi5Engine.recognizer_dispatch_name
    try:
        Dispatch(dispatch_name)
    except com_error as e:
        _log.exception("COM error during dispatch: %s", e)
        return False
    except Exception as e:
        _log.exception("Exception during Sapi5 COM dispatch): %s", e)
        return False
    return True


def get_engine(name=None, **kwargs):
    """
        Retrieve the Sapi5 back-end engine object.

        :param name: optional human-readable name of the engine to return.
        :type name: str
        :param \\**kwargs: optional keyword arguments passed through to the
            engine for engine-specific configuration.
    """
    global _engine
    if not _engine:
        from .engine import Sapi5Engine, Sapi5InProcEngine
        # Use the in-process engine by default.
        if not name or name == "sapi5inproc":
            _engine = Sapi5InProcEngine(**kwargs)
        else:
            _engine = Sapi5Engine(**kwargs)
    return _engine



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This file implements a Rectangle class for geometric operations.
"""

import copy
from dragonfly.windows.point import Point


#===========================================================================
# Rectangle class.

class Rectangle(Point):

    #-----------------------------------------------------------------------
    # Methods for initialization, copying, and introspection.

    def __init__(self, x1=None, y1=None, dx=None, dy=None):
        Point.__init__(self, x1, y1)
        self._size = Point(dx, dy)

    def copy(self):
        return copy.deepcopy(self)
    def __copy__(self):
        return copy.deepcopy(self)

    def __repr__(self):
        return "%s(%.1f, %.1f, %.1f, %.1f)" \
            % (self.__class__.__name__, self.x, self.y, self.dx, self.dy)

    def __eq__(self, other):
        return (super(Rectangle, self).__eq__(other) and
                self.size == other.size)

    def __ne__(self, other):
        return not self.__eq__(other)

    #-----------------------------------------------------------------------
    # Methods that control attribute access.

    p1 = property(fget=lambda self: Point(self._x, self._y),
                  doc="Protected access to p1 attribute.")
    p2 = property(fget=lambda self: self.p1 + self._size,
                  doc="Protected access to p2 attribute.")
    size = property(fget=lambda self: self._size.copy() ,
                    doc="Protected access to size attribute.")

    x1 = Point.x
    y1 = Point.y

    def _set_x2(self, x):
        self._size.x = x    # Use type checking of Point class.
        self._size._x -= self._x
    x2 = property(fget=lambda self: self._x + self._size._x,
                  fset=_set_x2,
                  doc="Protected access to x2 attribute.")

    def _set_y2(self, y):
        self._size.y = y    # Use type checking of Point class.
        self._size._y -= self._y
    y2 = property(fget=lambda self: self._y + self._size._y,
                  fset=_set_y2,
                  doc="Protected access to y2 attribute.")

    def _get_center(self):
        return self.p1 + Point(self._size.x / 2, self._size.y / 2)
    center = property(fget=_get_center,
                      doc="Dynamic access to center attribute.")

    dx = property(fget=lambda self: self._size.x,
                  doc="Protected access to dx attribute.")
    dy = property(fget=lambda self: self._size.y,
                  doc="Protected access to dy attribute.")

    x_center = property(fget=lambda self: self.x + self._size.x / 2,
                        doc="Protected access to x_center attribute.")
    y_center = property(fget=lambda self: self.y + self._size.y / 2,
                        doc="Protected access to y_center attribute.")

    ltwh = property(fget=lambda self: (int(self._x), int(self._y),
                                       int(self._size._x),
                                       int(self._size._y)),
                    doc="Shortcut to left-top-with-height tuple.")

    #-----------------------------------------------------------------------
    # Methods for manipulating rectangle objects.

    def translate(self, dx, dy):
        other = Point(dx, dy)
        self += other

    def renormalize(self, src, dst):
        # pylint: disable=W0212
        # Suppress warnings about protected member access.
        p1 = self.p1.renormalize(src, dst)
        p2 = self.p2.renormalize(src, dst)
        self._x = p1.x
        self._y = p1.y
        self._size._x = p2.x - p1.x
        self._size._y = p2.y - p1.y

    #-----------------------------------------------------------------------
    # Methods for various rectangle related operations.

    def contains(self, p):
        """Test whether this rectangle instance contains a point."""
        assert isinstance(p, Point)
        return self.x1 <= p.x < self.x2 and self.y1 <= p.y < self.y2


#===========================================================================
# Unity rectangle instance.

unit = Rectangle(0.0, 0.0, 1.0, 1.0)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

""" This file defines the base keyboard interface classes. """

import logging


class BaseKeySymbols(object):
    pass


class MockKeySymbols(BaseKeySymbols):
    def __getattribute__(self, _):
        # No keys can be typed, so return -1.
        return -1


class BaseKeyboard(object):
    """ Base keyboard interface. """

    _log = logging.getLogger("keyboard")

    @classmethod
    def send_keyboard_events(cls, events):
        """ Send a sequence of keyboard events. """
        # TODO Add a URL to the docs?
        raise NotImplementedError("Keyboard support is not implemented for "
                                  "this platform or your platform was not "
                                  "detected correctly.  Please see the "
                                  "documentation.")

    @classmethod
    def get_typeable(cls, char, is_text=False):
        """ Get a Typeable object. """
        return BaseTypeable(cls, char)


class BaseTypeable(object):
    """Container for keypress events."""

    __slots__ = ("_code", "_modifiers", "_name", "_is_text")

    _log = logging.getLogger("keyboard")

    def __init__(self, code, modifiers=(), name=None, is_text=False):
        """Set keypress information."""
        self._code = code
        self._modifiers = modifiers
        self._name = name
        self._is_text = is_text

    # pylint: disable=no-self-use,unused-argument
    def update(self, hardware_events_required):
        """
        Update keypress information.

        :rtype: bool
        :returns: success
        """
        return True

    def __repr__(self):
        """ Return information useful for debugging. """
        return ("%s(%s)" % (self.__class__.__name__, self._name) +
                repr(self.events()))

    def on_events(self, timeout=0):
        """Return events for pressing this key down."""
        events = [(m, True, 0) for m in self._modifiers]
        events.append((self._code, True, timeout))
        return events

    def off_events(self, timeout=0):
        """Return events for releasing this key."""
        events = [(m, False, 0) for m in self._modifiers]
        events.append((self._code, False, timeout))
        events.reverse()
        return events

    def events(self, timeout=0):
        """Return events for pressing and then releasing this key."""
        events = [(self._code, True, 0), (self._code, False, timeout)]
        for m in self._modifiers[-1::-1]:
            events.insert(0, (m, True, 0))
            events.append((m, False, 0))
        return events



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
EngineBase class
============================================================================

"""

import logging
from .timer import Timer

import dragonfly.engines


#---------------------------------------------------------------------------

class EngineError(Exception):
    pass

class MimicFailure(EngineError):
    pass


#---------------------------------------------------------------------------

class EngineContext(object):

    def __init__(self, engine):
        self._engine = engine

    def __enter__(self):
        self._engine.connect()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._engine.disconnect()


#---------------------------------------------------------------------------

class EngineBase(object):
    """ Base class for engine-specific back-ends. """

    _log = logging.getLogger("engine")
    _name = "base"
    _timer_manager = None

    #-----------------------------------------------------------------------

    def __init__(self):
        # Register initialization of this engine.
        dragonfly.engines.register_engine_init(self)

        self._grammar_wrappers = {}
        self._recognition_observer_manager = None

#    def __del__(self):
#        try:
#            try:
#                self.disconnect()
#            except Exception, e:
#                self._log.exception("Engine destructor raised an exception: %s" % e)
#        except:
#            pass

    def __repr__(self):
        return "%s()" % self.__class__.__name__

    @property
    def name(self):
        """ The human-readable name of this engine. """
        return self._name

    @property
    def grammars(self):
        """ Grammars loaded into this engine. """
        # Return a list of each GrammarWrapper's Grammar object.
        return list(map(
            lambda w: w.grammar,
            self._grammar_wrappers.values()
        ))

    def connect(self):
        """ Connect to back-end SR engine. """
        raise NotImplementedError("Virtual method not implemented for"
                                  " engine %s." % self)

    def disconnect(self):
        """ Disconnect from back-end SR engine. """
        raise NotImplementedError("Virtual method not implemented for"
                                  " engine %s." % self)

    def connection(self):
        """ Context manager for a connection to the back-end SR engine. """
        return EngineContext(self)

    #-----------------------------------------------------------------------
    # Methods for administrating timers.

    def create_timer(self, callback, interval, repeating=True):
        """ Create and return a timer using the specified callback and
            repeat interval. """
        return Timer(callback, interval, self._timer_manager, repeating)

    #-----------------------------------------------------------------------
    # Methods for administrating grammar wrappers.

    def _build_grammar_wrapper(self, grammar):
        raise NotImplementedError("Virtual method not implemented for"
                                  " engine %s." % self)

    #-----------------------------------------------------------------------
    # Methods for working with grammars.

    def load_grammar(self, grammar):
        wrapper_key = id(grammar)
        if wrapper_key in self._grammar_wrappers:
            self._log.warning("Grammar %s loaded multiple times." % grammar)
            return

        wrapper = self._load_grammar(grammar)
        self._grammar_wrappers[wrapper_key] = wrapper

    def _load_grammar(self, grammar):
        raise NotImplementedError("Virtual method not implemented for"
                                  " engine %s." % self)

    def unload_grammar(self, grammar):
        wrapper_key = id(grammar)
        if wrapper_key not in self._grammar_wrappers:
            raise EngineError("Grammar %s cannot be unloaded because"
                              " it was never loaded." % grammar)
        wrapper = self._grammar_wrappers.pop(wrapper_key)
        self._unload_grammar(grammar, wrapper)

    def _unload_grammar(self, grammar, wrapper):
        raise NotImplementedError("Virtual method not implemented for"
                                  " engine %s." % self)

    def update_list(self, lst, grammar):
        raise NotImplementedError("Virtual method not implemented for"
                                  " engine %s." % self)

    def activate_grammar(self, grammar):
        raise NotImplementedError("Virtual method not implemented for"
                                  " engine %s." % self)

    def deactivate_grammar(self, grammar):
        raise NotImplementedError("Virtual method not implemented for"
                                  " engine %s." % self)

    def activate_rule(self, rule, grammar):
        raise NotImplementedError("Virtual method not implemented for"
                                  " engine %s." % self)

    def deactivate_rule(self, rule, grammar):
        raise NotImplementedError("Virtual method not implemented for"
                                  " engine %s." % self)

    def set_exclusiveness(self, grammar, exclusive):
        """ Set the exclusiveness of a grammar. """
        raise NotImplementedError("Virtual method not implemented for"
                                  " engine %s." % self)

    def set_exclusive(self, grammar, exclusive):
        """ Alias of :meth:`set_exclusiveness`. """
        self.set_exclusiveness(grammar, exclusive)

    def _get_grammar_wrapper(self, grammar):
        wrapper_key = id(grammar)
        if wrapper_key not in self._grammar_wrappers:
            raise EngineError("Grammar %s never loaded." % grammar)
        wrapper = self._grammar_wrappers[wrapper_key]
        return wrapper

    #-----------------------------------------------------------------------
    # Recognition observer methods.

    def register_recognition_observer(self, observer):
        self._recognition_observer_manager.register(observer)

    def unregister_recognition_observer(self, observer):
        self._recognition_observer_manager.unregister(observer)

    def enable_recognition_observers(self):
        self._recognition_observer_manager.enable()

    def disable_recognition_observers(self):
        self._recognition_observer_manager.disable()


    #-----------------------------------------------------------------------
    #  Miscellaneous methods.

    def do_recognition(self, begin_callback=None, recognition_callback=None,
                       failure_callback=None, end_callback=None,
                       post_recognition_callback=None, *args, **kwargs):
        """
        Recognize speech in a loop until interrupted or :meth:`disconnect`
        is called.

        Recognition callback functions can optionally be registered.

        Extra positional and key word arguments are passed to
        :meth:`_do_recognition`.

        :param begin_callback: optional function to be called when speech
            starts.
        :type begin_callback: callable | None
        :param recognition_callback: optional function to be called on
            recognition success.
        :type recognition_callback: callable | None
        :param failure_callback: optional function to be called on
            recognition failure.
        :type failure_callback: callable | None
        :param end_callback: optional function to be called when speech
            ends, either successfully (after calling the recognition
            callback) or in failure (after calling the failure callback).
        :type end_callback: callable | None
        :param post_recognition_callback: optional function to be called
            after all rule processing has completed.
        :type post_recognition_callback: callable | None
        """
        # Import locally to avoid cycles.
        from dragonfly.grammar.recobs_callbacks import (
            register_beginning_callback, register_recognition_callback,
            register_failure_callback, register_ending_callback,
            register_post_recognition_callback
        )

        if begin_callback:
            register_beginning_callback(begin_callback)
        if recognition_callback:
            register_recognition_callback(recognition_callback)
        if failure_callback:
            register_failure_callback(failure_callback)
        if end_callback:
            register_ending_callback(end_callback)
        if post_recognition_callback:
            register_post_recognition_callback(post_recognition_callback)

        # Call _do_recognition() to start recognizing.
        self._do_recognition(*args, **kwargs)

    def _do_recognition(self, *args, **kwargs):
        """
        Recognize speech in a loop until interrupted or :meth:`disconnect`
        is called.

        This method should be implemented by engine sub-classes.
        """
        raise NotImplementedError("Engine %s not implemented." % self)

    #: Alias of :meth:`do_recognition` left in for backwards-compatibility
    recognize_forever = do_recognition

    #: Alias of :meth:`do_recognition` left in for backwards-compatibility
    recognise_forever = do_recognition

    def process_grammars_context(self, window=None):
        """
            Enable/disable grammars & rules based on their current contexts.

            This must be done preemptively for some SR engine back-ends,
            such as WSR, that don't apply context changes upon/after the
            utterance start has been detected. The WSR engine should call
            this automatically whenever the foreground application (or its
            title) changes. The user may want to call this manually to
            update when custom contexts.

            The *window* parameter is optional window information, which can
            be passed in as an optimization if it has already been gathered.

        """

        if window is None:
            from dragonfly.windows.window import Window
            window = Window.get_foreground()
        for grammar in self.grammars:
            # Prevent 'notify_begin()' from being called.
            if grammar.name == "_recobs_grammar":
                continue
            grammar.process_begin(window.executable, window.title,
                                  window.handle)

    def mimic(self, words):
        """ Mimic a recognition of the given *words*. """
        raise NotImplementedError("Engine %s not implemented." % self)

    def speak(self, text):
        """ Speak the given *text* using text-to-speech. """
        raise NotImplementedError("Engine %s not implemented." % self)

    @property
    def language(self):
        """
        Current user language of the SR engine. (Read-only)

        :rtype: str
        """
        return self._get_language()

    def _get_language_tag(self, language_id):
        # Get a language tag from the Windows language identifier.
        tags = self._language_tags.get(language_id)
        if tags:
            return tags[0]

        # The _language_tags dictionary didn't contain the language, so
        # get the best match by using the primary language identifier.
        # This allows us to match unlisted language variants.
        primary_id = language_id & 0x00ff
        for language_id2, (tag, _) in self._language_tags.items():
            if primary_id == language_id2 & 0x00ff:  # Match found.
                return tag

        # Speaker language wasn't found.
        self._log.error("Unknown speaker language: 0x%04x" % language_id)
        raise EngineError("Unknown speaker language: 0x%04x" % language_id)

    _language_tags = {
                      0x0c09: ("en", "AustralianEnglish"),
                      0xf00a: ("es", "CastilianSpanish"),
                      0xf809: ("en", "CAEnglish"),
                      0x0004: ("zh", "Chinese"),
                      0x0413: ("nl", "Dutch"),
                      0x0009: ("en", "English"),
                      0x040c: ("fr", "French"),
                      0x0407: ("de", "German"),
                      0xf009: ("en", "IndianEnglish"),
                      0x0410: ("it", "Italian"),
                      0x0411: ("jp", "Japanese"),
                      0xf40a: ("es", "LatinAmericanSpanish"),
                      0x0416: ("pt", "Portuguese"),
                      0xf409: ("en", "SingaporeanEnglish"),
                      0x040a: ("es", "Spanish"),
                      0x0809: ("en", "UKEnglish"),
                      0x0409: ("en", "USEnglish"),
                     }

    def _get_language(self):
        raise NotImplementedError("Engine %s not implemented." % self)

    @property
    def quoted_words_support(self):
        """
        Whether this engine can compile and recognize quoted words.

        :rtype: bool
        """
        return self._has_quoted_words_support()

    def _has_quoted_words_support(self):
        raise NotImplementedError("Engine %s not implemented." % self)
